# Rest API Layer for On-Demand Complaint Timeline and Trends
# Show timeline on demand
# produce a timeline showing the complaint counts for a query
# if there is a trend in the timeline -> show the trend start date

"""
-----------------------------------
---------Main import libraries------
-----------------------------------
"""

# imports
from trendDetection_func import *
import sys
import json
import argparse
import logging
from logging.handlers import TimedRotatingFileHandler
from logging import getLogger
import os
import ibm_db
import ibm_db_dbi
import ssl
import pandas as pd
from flask import Flask, jsonify, request
from flask_cors import CORS
import configparser

# create flask app
app = Flask(__name__)
APP_ROOT = os.path.dirname(os.path.abspath(__file__))
CORS(app)

# the rest API can be called with optional command line parameters for testing and debugging
parser = argparse.ArgumentParser()
parser.add_argument("-gc", "--global-sifs-config-file", dest="sifs_config_file",
                    default='/home/sifsuser/lib/sifs.spark.properties', help="global SIFS configuration file")
parser.add_argument("-rc", "--rest-config-file", dest="rest_config_file_name", default='REST.cfg',
                    help="REST API configuration file")
parser.add_argument("-tc", "--trends-config-file", dest="trends_config_file_name", default='../config/config.cfg',
                    help="trends models configuration file")
parser.add_argument("-lg", "--log-file", dest="log_file_name", default='../config/logs/on_demand_trends.log',
                    help="log file")
parser.add_argument("-c", "--cert-file", dest="cert_file_name", help="python ssl to db2 ARM public certificate")
parser.add_argument("-nv", "--no_ssl_verification", dest="no_ssl_verification", action="store_true")
args = vars(parser.parse_args())

# Read Config files
TRENDS_HOME = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
config_path = os.path.join(TRENDS_HOME, 'config')
config = configparser.ConfigParser()
config_file = os.path.join(config_path, args['rest_config_file_name'])
cfg = config.read(config_file)

if not len(cfg):
    txt = 'Unable to read https config file {} or blank config file given. Ending Script.'.format(config_file)
    # logging.error(txt) logging hasn't been set up yet... it needs to read config first eg logFileName
    quit()

# # logging
if not os.path.isdir(config_path + "/logs/"):
    os.makedirs(config_path + "/logs/")
config_file_name = os.path.join(config_path, args['log_file_name'])

log_level = 10
handler = TimedRotatingFileHandler(config_file_name, when="midnight", interval=1)
handler.setLevel(log_level)

log_format = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
formatter = logging.Formatter(log_format)
handler.setFormatter(formatter)

# add a suffix which you want
handler.suffix = "%Y%m%d" + str(".log")
logging.basicConfig(level=logging.DEBUG, handlers=[handler])

logger = getLogger(os.path.basename(__file__))
# this echoes logger commands to stdout so that app status can be seen on screen
logging.basicConfig(format=log_format)
logging.getLogger().addHandler(logging.StreamHandler())

# get https config details: debug flag and certificate, key paths
port = config['server']['Port']
serverDebug = config['server']['Debug'] == 'True'  # convert to bool
certPath = config['server']['certPath']
keyPath = config['server']['keyPath']

# use param override for testing e.g C:\\Users\\IBM_ADMIN\\.spyder-py3\\SIDB2.arm;"
sslServerCert = args['cert_file_name']
if not sslServerCert:
    sslServerCert = config['server']['sslServerCert']  # path to ARM cert

logFileName = config['log_file']['LogFileName']


def jsonifys(status_code=200, **kwargs):
    """ turn a response into json
    """
    response = jsonify(**kwargs)
    response.status_code = status_code
    return response

# get all configuration parameters
try:
    db2_params, trends_config = get_config(args['sifs_config_file'],
                                           os.path.join(config_path, args['trends_config_file_name']),
                                           sslServerCert,
                                           logger)
except ValueError as err:
    logger.error('Error getting config', exc_info=True)
    print('Error unable to read config: {}'.format(err.args))
    quit()

logger.info('SUCCESS: All Config files read.')

#  Main api functions
# on-demand trend
@app.route('/analytics/models/v1/on_demand_trend/', methods=['POST'])
def on_demand_trend():

    # check that the query is contained in the body
    body = request.get_json()
    query = body.get('query')

    if query is None:
        logger.error('Missing parameter: query', exc_info=True)
        return jsonifys(500, error="Missing parameter 'query' in json request.")

    logger.info('SUCCESS: read on-demand query: {}'.format(query))

    # establish connection to db2
    logger.info('connecting to host {} port {}'.format(db2_params['host'], db2_params['port']))
    try:
        ibm_db_conn = ibm_db.pconnect(db2_params['conn_str'], '', '')
        connection = ibm_db_dbi.Connection(ibm_db_conn)
        connection.set_current_schema(db2_params['database'])
        logger.info('connected to db2')
    except:
        logger.error('Unable to connect to db2')
        return jsonifys(500, error="Unable to connect to db2")

    # call the on-demand trend function
    try:
        trends = get_trends(query, connection, trends_config, logger)
    except ValueError as er:
        txt = 'Error finding trend(s) for query {}'.format(query)
        logger.error(txt)
        print(txt)
        return jsonifys(500, error=txt)

    logging.info('On-demand trend created: {}'.format(trends))
    return json.dumps({'themedetails': trends})

if __name__ == "__main__":
    context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    cer = os.path.join(config_path, certPath)
    key = os.path.join(config_path, keyPath)

    certKey = True
    if not os.path.isfile(cer):
        certKey = False
        logger.error('Error finding certificate: ' + cer)
        print('Missing certificate file', cer)

    if certKey:
        context.load_cert_chain(cer, key)
        app.run(host='0.0.0.0', port=int(port), debug=serverDebug, ssl_context=context, threaded=True)
    else:
        logger.error('Unable to start server due to missing certificate and/or key')
        print('Unable to start server due to missing certificate and/or key. Check config file and cert/key files.')
