# coding: utf-8

# Show timeline on demand
# This is a standalone python script for on-demand trends
# DOES NOT RUN AS A REST API
# produce a timeline showing the complaint counts for a query
# if there is a trend in the timeline -> show the trend start date

# imports
from trendDetection_func import *
import sys
import json
import argparse
import logging
from logging.handlers import TimedRotatingFileHandler
from logging import getLogger
import os
import re
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
import ibm_db
import ibm_db_dbi

# config parser
try:
    import configparser as configparser
except ImportError:
    import configparser
config = configparser.ConfigParser()

# parse command line arguments
parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input-file", dest="input_file_name", help="read input data from csv file")
parser.add_argument("-db2", "--use_db2", dest="db2", action="store_true", help="read input data from db2")
parser.add_argument("-o", "--output_file", dest="output_file_name", help='save output data to csv file')
parser.add_argument("-r", "--rest_api", dest="rest_api", action="store_true", help='send output to rest api')
parser.add_argument("-lw", "--long_window_length", dest="long_window", help="Length in days of window used\
                                                                                to determine the normal variability")
parser.add_argument("-sw", "--short_window_length", dest="short_window", help="Has a trend been detected in\
                                                                                    the last x days ?")
parser.add_argument("-c", "--config-file", dest="config_file_name", default='../config/config.cfg',
                    help="configuration file")
parser.add_argument("-lg", "--log-file", dest="log_file_name", default='../config/logs/on_demand_trends.log',
                    help="configuration file")
parser.add_argument("-nv", "--no_ssl_verification", dest="no_ssl_verification", action="store_true")
parser.add_argument("-t", "--test", dest="test", action="store_true", help="read a local copy of spark.sifs.config")
args = parser.parse_args()

# logging
ANALYTICS_HOME = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
config_path = os.path.join(ANALYTICS_HOME, "config")
log_path = os.path.join(config_path, "logs")
if not os.path.isdir(log_path):
    os.makedirs(log_path)

log_file_name = os.path.join(log_path, args.log_file_name)
log_level = logging.INFO
handler = TimedRotatingFileHandler(args.log_file_name, when="midnight", interval=1)
handler.setLevel(log_level)

log_format = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
formatter = logging.Formatter(log_format)
handler.setFormatter(formatter)
handler.suffix = "%Y%m%d" + str(".log")

logger = getLogger(os.path.basename(__file__))
# this echoes logger commands to stdout so that app status can be seen on screen
logging.basicConfig(level=log_level, format=log_format, handlers=[handler])
logging.getLogger().addHandler(logging.StreamHandler())

# check python version
if sys.version_info < (3, 0):
    logging.error('Must be run with python version 3 or above. Ending Script.')
    quit()

if 'args' not in locals():
    logging.error('No command line parameters passed. Ending Script.')
    quit()

# bool flag for whether to read data from db2, send data via REST API
db2 = args.db2
rest_api = args.rest_api
test = args.test

# check input file combinations
if args.input_file_name is None and not db2:
    input_file_name = None
    logging.error('No input data parameters: choose either -db2 or input file -i.')
    quit()
else:
    input_file_name = args.input_file_name
    if 'hdfs' in args.input_file_name:
        pass
    else:
        if not os.path.isfile(input_file_name):
            txt = 'Unable to find local data file', input_file_name, '. Ending Script.'
            logging.error(txt)
            quit()

# read global SIFS config: check the spark.sifs.complaints config
if test: # test mode - use a local file
    spark_sifs_config = '../config/sifs.spark.properties'
else: # use the SIFS standard path
    spark_sifs_config = '/home/sifsuser/lib/sifs.spark.properties'

if rest_api or db2:
    # todo update param
    # param = "ComplaintTrendsRESTonDemand"
    on_demand_url = "ComplaintTrendsREST"

    if not os.path.isfile(spark_sifs_config):
        txt = 'Unable to find global SIFS config file: ' + spark_sifs_config + '. Update the file path or use parameter' \
              '-t to read local file ../config/sifs.spark.properties.  Ending Script.'
        logging.error(txt)
        quit()

    f_sifs = open(spark_sifs_config)
    lines = f_sifs.readlines()

    if rest_api:
        url_line = [l for l in lines if on_demand_url in l]
        if url_line:
            _, api_url_raw = url_line[0].split('=')
            api_url = api_url_raw.replace('\n', '').strip()
            txt = 'Success: read ComplaintTrendsREST paramater in config file:' + api_url
            logging.info(txt)
        else:
            txt = 'Unable to find {} paramater in config file. Ending Script.'.format(on_demand_url)
            logging.error(txt)
            quit()

    # read db2 integration params
    if db2:
        db2_params = {'db2user': '', 'db2password': '', 'db2jdbcurl': ''}
        for p in db2_params.keys():
            param = [l for l in lines if str(p) in l]
            if param:
                p_name, p_val = param[0].split('=')[0], param[0].split('=')[1]
                db2_params[p] = p_val.replace('\n', '').strip()
                logging.info('Read parameter {} from config file.'.format(p))
            else:
                logging.error('Unable to find parameter {} in config file. Ending Script.'.format(p))
                quit()

        # # jdbc url is in the format jdbc:db2://ip:port/SIFS: only want ip
        # if ':sslConnection' in db2_params['db2jdbcurl']:
        #     url = db2_params['db2jdbcurl'].split(':sslConnection')[0]
            # to do some issue with wrong port set up in sifs.config,
            if '50001' in db2_params['db2jdbcurl']:
                db2_params['db2jdbcurl'] = db2_params['db2jdbcurl'].replace('50001', '50000')

# check output destination combinations
# bool flag for whether to send data using REST API to db2
rest_api = args.rest_api
if args.output_file_name is None and not rest_api:
    output_file_name = None
    logging.error('No destination for output data: choose either -r for REST API or -o for output file.')
    quit()
else:
    if rest_api:
        logging.info('output will be sent using REST API to db2')
    if args.output_file_name is not None:
        if isinstance(args.output_file_name, str):
            output_file_name = args.output_file_name
            # validate the file name
            p_directory, fname = os.path.split(output_file_name)
            match = re.match('(\w+)\.csv$', fname)
            if match is None:
                txt = 'Invalid output file parameter: ', output_file_name, '.Should be "xxx.csv". Ending Script.'
                logging.error(txt)
                quit()

            # check directory
            out_folder = '../output/'
            if not p_directory:
                txt = 'Pre-pending output directory for output file: ' + out_folder
                logging.info(txt)
                output_file_name = os.path.join(out_folder, output_file_name)
            else:
                # overwrite any other file path
                output_file_name = os.path.join(out_folder, output_file_name)

            if os.path.isfile(output_file_name):
                txt = 'Output file already exists and will be overwritten: ' + output_file_name
                logging.info(txt)
                # check if file is already open
                try:
                    myfile = open(output_file_name, "r+")
                    myfile.close()
                except IOError:
                    txt = 'Output file: ' + output_file_name + ' is already open. Please close this file and re-run.' \
                                                                'Ending Script.'
                    logging.error(txt)
                    quit()

        else:
            txt = 'Invalid local file parameter: ' + args.output_file_name + '. Ending Script.'
            logging.error(txt)
            quit()

# check config file
if args.config_file_name is None:
    logging.error('No config file paramater given. Ending Script.')
    quit()
else:
    # read config file
    if args.config_file_name is None:
        txt = 'Unable to read config file or blank config file given. Ending Script.'
        logging.error(txt)
        quit()
    else:
        cfg = config.read(os.path.join(config_path, args.config_file_name))

        if not len(cfg):
            txt = 'Unable to read config file or blank config file given. Ending Script.' + args.config_file_name
            logging.error(txt)
            quit()
        else:
            logging.info('Config files read completed')

# check short window param
if args.short_window is not None:
    short_window = int(args.short_window)
else:
    logging.info('Parameter "short window": using default value 7 days.')
    short_window = 7

# check long window param
if args.long_window is not None:
    long_window = int(args.long_window)
else:
    logging.info('Parameter "long window": using default value 30 days.')
    long_window = 30

# ssl verification, switch to false for debug only
verify = True
if args.no_ssl_verification is not None:
    verify = False

# get db2 integration params from config file
if config is not None:
    database = config.get("db2 integration", "database")  # database = SIFS
    table = config.get("db2 integration", "table")        # table = SIFS.COMPLAINT_FEATURES

# check the config
if config is not None:
    mk_confidence = config.getfloat("mann kendall", "mk_confidence")
    min_records = config.getint("trend candidates", "min_records")
    min_per_day = config.getint("trend candidates", "min_per_day")
    pct_change_threshold = config.getfloat("trend candidates", "pct_change_threshold")
    MIN_MONOTONE = config.getfloat("trend candidates", "min_monotone")
    fields_list = config.get("trend candidates", "fields_list")
    fields_list = fields_list.replace('-', '_').split(',')

    # check for filter on products
    products = config.get("trend candidates", "products")
    if products == 'all':
        products = []
    else:
        products = products.split(",")
    model_config = dict(config.items('poisson model'))
    model_name = config.get('poisson model', "model_name")

    # v2 updates
    MAX_RISK = config.getfloat("risk score", "max_risk")  # 85% # 503 max risk score allowable is 85%
    RISK_WEIGHTS = json.loads(config.get("risk score", "risk_weights"))  # rel. weights of factors for scoring trends
    MAX_COUNT = config.getfloat("risk score", "max_count")
    MAX_GRAD = config.getfloat("risk score", "max_grad")
    MIN_RISK_FILTER = config.getfloat("risk score", "min_risk_filter")

    # on-demand params
    # number of days for averaging the raw complaint count. e.g. use 7 day average to remove weekly seasonality
    smoothfactor = config.getint("on-demand", "smoothfactor")
    # trend on days that are less than this factor apart will be joined up to find the earliest date the trend started
    DAYS_BETWEEN_TRENDS = config.getint("on-demand", "DAYS_BETWEEN_TRENDS")
    # number of time periods to calculate the on-demand timelines
    time_periods = json.loads(config.get("on-demand", "time_periods"))    # e.g. [365, 90, 30, 7]

    # check validity of risk weights
    if len(RISK_WEIGHTS) != 3:
        # warning default equal weighting used for all weights
        txt = 'Default values are used to give equal weighting to all factors in the risk score calculation.'
        logging.info(txt)
        RISK_WEIGHTS = [1 / 3 for _ in range(3)]
    # normalise weights if required
    if sum(RISK_WEIGHTS) != 1:
        RISK_WEIGHTS = np.asarray(RISK_WEIGHTS) / sum(RISK_WEIGHTS)
else:
    txt = 'Unable to read config file. Ending Script.'
    logging.error(txt)
    quit()

keep_cols = ['complaintcount', 'description', 'geo', 'product', 'riskscore', 'theme', 'timeperiod',
             'trendattributes', 'trendstartdate', 'trendtypeid']

# initialise a poisson trend detection model
model = getattr(td_models, model_name)(config=model_config)

# Main processing starts here
host = re.search('([0-9]{1,3}\.){3}([0-9]{1,3})', db2_params['db2jdbcurl']).group()   # get host ip from rest url
port = re.search('(?![0-9]\:)[0-9]{5}', db2_params['db2jdbcurl']).group()             # get host port from rest url

conn_str = "DATABASE=%s;HOSTNAME=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s" % (database,
                                                                             host,
                                                                             port,
                                                                             db2_params['db2user'],
                                                                             db2_params['db2password'])
logging.info('connecting to host {} port {}'.format(host, port))
try:
    ibm_db_conn = ibm_db.connect(conn_str, '', '')
    conn = ibm_db_dbi.Connection(ibm_db_conn)
    conn.set_current_schema(database)
    logging.info('connected to db2')
except:
    logging.error('Unable to connect to db2')
    exit()

# dynamically build up the query
query = {"THEME": ["Complaint_SalesPractice", "Complaint_Delay"], 'CUSTOMER_AGE': ['30-40', '40-50'], 'GEO': ['TX','FL']}  # chosen by user in UI
logging.info('evaluate UI query: {}'.format(json.dumps(query)))

# check for REF fields:
#  its possible that GEO is a being queried but geo data is actually stored in the GEO REF column
sql_query = query.copy()  # sql_query is used to build the query, query is used for outputting texts etc
sql_query.update({'IS_COMPLAINT': ['TRUE']})  # we only want complaints, so add this to where clause in sql query
ref_fields = ['PRODUCT', 'SUB_PRODUCT', 'GEO']
replace_fields = [rf for rf in ref_fields if (rf in query) and (rf+'_REF' in fields_list)]
for rf in replace_fields:
    replace_field = rf.replace('_REF', '')
    sql_query[rf+'_REF'] = query[rf]
    del sql_query[rf]

if replace_fields:
    logging.info('query updated to use _REF fields: {}'.format(json.dumps(sql_query)))

# get start date and end date
end_date = query.get('ENDDATE', None)
if end_date:
    if valid_date(end_date):
        logging.info('end date parameter: {}'.format(end_date))
    else:
        logging.error('Invalid end date parameter: {}. Ending Script.'.format(end_date))
        quit()
else:
    logging.info('end date: Parameter "ENDDATE" not in query, using todays date.')
    end_date = datetime.now().strftime("%Y-%m-%d")
dt_end_date = datetime.strptime(end_date, '%Y-%m-%d')

start_date = query.get('STARTDATE', None)
if start_date:
    if valid_date(start_date):
        logging.info('start date parameter: {}'.format(start_date))
    else:
        logging.error('Invalid start date parameter: {}. Ending Script.'.format(start_date))
        quit()
else:
    logging.info('start date: start today, go back {} days + use prior {} days to get context'
                 .format(max(time_periods) - 1, long_window))
    # the start date is the greatest number of days in time periods + the number of days in the long window
    start_date = datetime.strftime(dt_end_date - timedelta(days=long_window + max(time_periods) - 1), '%Y-%m-%d')
dt_start_date = datetime.strptime(start_date, '%Y-%m-%d')

logging.info('on-demand date range: start date={}'.format(start_date))
logging.info('on-demand date range: end   date={}'.format(end_date))

# dynamically build up the query
select_fields = ["TO_DATE(COMPLAINT_DATE,'MM-DD-YYYY') as date", 'COUNT(ID) as count']  # need these to build a timeline
select_fields_s = ' ,'.join(select_fields)

where_s = get_where_clause(sql_query, start_date, end_date)

qry = 'select {} from sifs.complaint_features where {} group by COMPLAINT_DATE'.format(select_fields_s, where_s)
logging.info('created db2 query: {}'.format(json.dumps(qry)))

# run the query
logging.info('started reading the data')
try:
    df_data = pd.read_sql(qry, conn, index_col='DATE')
except:
    print("error running query: '{}'".format(qry))
logging.info('finished reading the data')

od_index = pd.date_range(start=start_date, end=end_date, freq='1D', name='date')
df_data = df_data.reindex(index=od_index, fill_value=0)
df_data.rename(columns={'COUNT': 'actualcount'}, inplace=True)
df_data['smoothcount'] = np.round(df_data['actualcount'].rolling(window=smoothfactor).mean(), 2)
df_data.fillna(0, inplace=True)
od_raw_counts, od_smooth_counts =  list(df_data['actualcount']), list(df_data['smoothcount'])

# set defaults and initialisations
summary = None
trends= []
trendstartdate = None

for day in od_index[long_window:]:  # run a sliding window of length <<long_window>> over the whole period
    # index of "to" date, add a +1 because want results to be inclusive of this end date
    idx_to = list(od_index).index(day) + 1
    # index of "from" date
    idx_from = idx_to - long_window
    lw_index = od_index[idx_from: idx_to]
    # return one row per trend discovered
    trend = od_get_trends (od_raw_counts[idx_from: idx_to], od_smooth_counts[idx_from: idx_to],
                                 od_index[idx_from: idx_to], short_window, long_window, MIN_MONOTONE, model,
                                 mk_confidence, float(model_config['poisson_threshold']), pct_change_threshold,
                                 MAX_RISK, MAX_COUNT, RISK_WEIGHTS, MAX_GRAD)
    if trend:
        trends.append(trend)

    # derive a minimal set of timeline statistics
    # if no trends have been found, still show trending 30 days etc so user can see "no trend" is the correct analysis
    if day == od_index[-1]:
        summary_min = get_min_stats(od_raw_counts[idx_from: idx_to], od_index[idx_from: idx_to], short_window, long_window)

# only save all trends to csv for debugging
# pd.DataFrame(pd.DataFrame(trends)).to_csv('../output/od_all_trends.csv', index=False)

# trend details: use the trend summary from the end date
for trend in sorted(trends, key=lambda x: x['key_date'], reverse=True):
    if trend['key_date'] == end_date:
        summary = trend
        trendstartdate = trend['trendstartdate']
        key_date = end_date

    elif trendstartdate:  # if trend startdate already set
        date_diff = datetime.strptime(key_date, '%Y-%m-%d') - datetime.strptime(trend['key_date'], '%Y-%m-%d')
        if date_diff.days <= DAYS_BETWEEN_TRENDS:
            trendstartdate = trend['trendstartdate']  # overwrite with earlier trend start date
            key_date = trend['key_date']

# if no trend has been found then give a minimal set of timeline stats
if not summary:
    summary = summary_min

description = "&".join([k.replace('&', '_') + ':' + ','.join([str(v).replace('&', '_')
                                                               for v in vs]) for k, vs in query.items()])

# create dataframe with common values
df_trends_od = pd.DataFrame({'selection_criteria': [query for _ in time_periods],
                             'timeperiod': sorted(time_periods, reverse=True),
                             'description': [description for _ in time_periods],
                             'product': [', '.join(query.get('PRODUCT', ['all'])) for _ in time_periods],
                             'trendattributes': [map_selections_on_demand(query) for _ in time_periods],
                             'geo': [','.join(query.get('GEO', 'all')) for _ in time_periods],
                             'theme': [', '.join(query.get('THEME', 'all')) for _ in time_periods],
                             'summary': [summary for _ in time_periods],
                             'trendstartdate': [trendstartdate for _ in time_periods]})

df_trends_od['datecounts'] = df_trends_od['timeperiod'].apply(od_datecounts, args=(od_raw_counts,
                                                                                       od_index, short_window))

df_trends_od['date_range'] = df_trends_od['timeperiod'].apply(lambda x: datetime.strftime(dt_end_date -
                                                                                          timedelta(days=x),
                                                                                          '%Y-%m-%d')+':'+end_date)

df_trends_od["riskscore"] = df_trends_od['summary'].map(lambda x: x.get('riskscore', 0.0))
df_trends_od["trendtypeid"] = df_trends_od['summary'].map(lambda x: x.get('trendtypeid', 3))  # default = 3 no trend
df_trends_od = df_trends_od.apply(create_complaint_count, axis=1)

#  append results to output structure
od_trends = df_trends_od[keep_cols].to_dict(orient='records')

# add final summary to log
logging.info('Found {} trends'.format(len(od_trends)))
for t in od_trends:
    logging.info(json.dumps(t))

# write results to local csv for testing and debugging
if len(od_trends) and output_file_name is not None:
    logging.info('saving on-demand trends to local file {}'.format(output_file_name))
    df_trends_od[keep_cols].to_csv(output_file_name, index=False)

# send the results by REST api to db2 database
if len(od_trends) and rest_api:
    logging.info('upload trends using REST api')
    response = post_request(od_trends, api_url, verify)
    logging.info("response status code:{}".format(response.status_code))
    logging.info("response status message: {}".format(response.text))
    if response.status_code != 200:
        logging.error('db update error')

# close
logging.info('All done. Ending script.')
# reset logging level to avoid errors after 'tear-down'
log_level = logging.INFO
handler.setLevel(log_level)
logging.basicConfig(level=log_level, format=log_format, handlers=[handler])
