# coding: utf-8

# Detect Trend Combinations
# This script identifies combinations of dimensions( product, issue, state, gender etc)  that could have be trending

# imports
from trendDetection_func import *
import sys
import csv
import json
import argparse
import logging
from logging.handlers import TimedRotatingFileHandler
from logging import getLogger
import os
import re
import numpy as np
import pandas as pd

# imports for detect trend  poisson model
from datetime import datetime, timedelta

from pyspark import SparkConf, SparkContext
from pyspark.sql import SparkSession
from pyspark.sql.types import StructField, StructType, StringType, LongType, DateType, BooleanType
from pyspark.sql import SQLContext
from pyspark.sql.functions import col, udf
from pyspark.sql.functions import datediff, months_between, to_date
from pyspark.sql.types import *
from pyspark.sql import Row

# config parser
try:
    import configparser as configparser
except ImportError:
    import configparser
config = configparser.ConfigParser()

# parse command line arguments
parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input-file", dest="input_file_name", default=None)
parser.add_argument("-sd", "--start_date", dest="start_date", help="start_date")
parser.add_argument("-ed", "--end_date", dest="end_date", help="end_date")
parser.add_argument("-lw", "--long_window_length", dest="long_window", help="Length in days of window used\
                                                                                to determine the normal variability")
parser.add_argument("-sw", "--short_window_length", dest="short_window", help="Has a trend been detected in\
                                                                                    the last x days ?")
parser.add_argument("-c", "--config-file", dest="config_file_name", default='../config/config.cfg',
                    help="configuration file")
parser.add_argument("-lg", "--log-file", dest="log_file_name", default='../config/logs/trend_detection.log',
                    help="configuration file")
parser.add_argument("-v", "--verbose", dest="verbose", help="increase output verbosity", action="store_true")
parser.add_argument("-s", "--spark", dest="spark_flag", action="store_true")  # if -s = True then spark else python
parser.add_argument("-lo", "--local_output_file", dest="output_file_name",
                    help='save trends to csv on local file system')
parser.add_argument("-z", "--zipcodemap", dest="zipcode", help="map zipcodes to State", action="store_true")
parser.add_argument("-nv", "--no_ssl_verification", dest="no_ssl_verification", action="store_true")
parser.add_argument("-db2", "--use_db2", dest="db2", action="store_true")
parser.add_argument("-t", "--test", dest="test", action="store_true", help="read a local copy of spark.sifs.config")
args = parser.parse_args()

# logging
ANALYTICS_HOME = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
config_path = os.path.join(ANALYTICS_HOME, "config")
log_path = os.path.join(config_path, "logs")
if not os.path.isdir(log_path):
    os.makedirs(log_path)

log_file_name = os.path.join(log_path, args.log_file_name)
log_level = logging.INFO
handler = TimedRotatingFileHandler(args.log_file_name, when="midnight", interval=1)
handler.setLevel(log_level)

log_format = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
formatter = logging.Formatter(log_format)
handler.setFormatter(formatter)
handler.suffix = "%Y%m%d" + str(".log")

logger = getLogger(os.path.basename(__file__))
# this echoes logger commands to stdout so that app status can be seen on screen
logging.basicConfig(level=log_level, format=log_format, handlers=[handler])
logging.getLogger().addHandler(logging.StreamHandler())

# check python version
if sys.version_info < (3, 0):
    txt = 'The script has been called with python version 2 but requires python version 3. Please run script again \
         using either python3 or update the PYSPARK_PYTHON system parameter.  Ending Script.'
    logging.error(txt)
    quit()

db2 = args.db2
test = args.test

# read system parameters for integration
if test:  # test mode - use a local file
    spark_sifs_config = '../config/sifs.spark.properties'
else:  # use the SIFS standard path
    spark_sifs_config = '/home/sifsuser/lib/sifs.spark.properties'

if args.output_file_name is None or db2:
    if not os.path.isfile(spark_sifs_config):
        txt = 'Unable to find global SIFS config file: ' + spark_sifs_config + '. Update file path or use parameter' \
              ' -t to read local file ../config/sifs.spark.properties.  Ending Script.'
        logging.error(txt)
        quit()

    # read global SIFS config: check the spark.sifs.complaints config
    rest_url_param = "ComplaintTrendsREST"
    # API url ComplaintTrendsREST = https://ip:port/complaintsservices/surveillance/v1/complaint/createTrend
    f_sifs = open(spark_sifs_config)
    lines = f_sifs.readlines()

    url_line = [l for l in lines if rest_url_param in l]
    if url_line:
        _, api_url_raw = url_line[0].split('=')
        api_url = api_url_raw.replace('\n', '').strip()
        txt = 'Success: read ComplaintTrendsREST parameter in config file:' + api_url
        logging.info(txt)
    else:
        txt = 'Unable to find ComplaintTrendsREST parameter in config file. Ending Script.'
        logging.error(txt)
        quit()

    if db2:  # data will be read from db2
        # read db2 integration params
        lines = [l for l in lines if l[0] != '#']
        db2_params = {'db2user': '', 'db2password': '', 'db2jdbcurl': '', 'db2TrustStore': ''}
        for p in db2_params.keys():
            param = [l for l in lines if str(p) in l]
            if param:
                p_name, p_val = param[0].split('=')[0], '='.join(param[0].split('=')[1:])
                db2_params[p] = p_val.replace('\n', '').strip()
                logging.info('Read parameter {} from config file.'.format(p))
            else:
                logging.error('Unable to find parameter {} in config file. Ending Script.'.format(p))
                quit()

        # change this to decode from base64
        if db2_params['db2password'].__contains__('{xor}'):
            db2_params['db2password'] = db2_params['db2password'][5:]
            db2_params['db2password'] = base64.b64decode(db2_params['db2password'])
            db2_params['db2password'] = bytearray(i ^ ord('_') for i in db2_params['db2password'])
            db2_params['db2password'] = db2_params['db2password'].decode('utf-8')
            logging.info('password decoded.')
        else:
            logging.error("XOR-base64 Encoding not found in config")
            quit()

        user = db2_params['db2user']
        passwd = db2_params['db2password']
        jdbcURL = db2_params['db2jdbcurl']
        tsl = db2_params['db2TrustStore']

# check config file
if args.config_file_name is None:
    logging.error('No config file parameter given. Ending Script.')
    quit()
else:
    # read config file
    if args.config_file_name is None:
        txt = 'Unable to read config file or blank config file given. Ending Script.'
        logging.error(txt)
        quit()
    else:
        cfg = config.read(os.path.join(config_path, args.config_file_name))

        if not len(cfg):
            txt = 'Unable to read config file or blank config file given. Ending Script.' + args.config_file_name
            logging.error(txt)
            quit()
        else:
            logging.info('SUCCESS: Config files read completed')

# check input file
if args.input_file_name is None and not db2:
    input_file_name = None
    logging.error('No input data parameters: choose either -db2 or input file -i.')
    quit()
else:
    if args.input_file_name is not None:
        input_file_name = args.input_file_name
        if 'hdfs' in args.input_file_name:
            pass
        else:
            if not os.path.isfile(input_file_name):
                txt = 'Unable to find local data file', input_file_name, '. Ending Script.'
                logging.error(txt)
                quit()

# check short window param
if args.short_window is not None:
    short_window = int(args.short_window)
else:
    txt = 'Parameter "short window": using default value 7 days.'
    logging.info(txt)
    short_window = 7

# check long window param
if args.long_window is not None:
    long_window = int(args.long_window)
else:
    txt = 'Parameter "long window": using default value 30 days.'
    logging.info(txt)
    long_window = 30

if args.start_date is not None:
    startDate = args.start_date
else:
    txt = 'Parameter "start date": using default value todays date minus', str(long_window)
    logging.info(txt)
    endDate = datetime.now().strftime("%Y-%m-%d")
    startDate = (datetime.now() - timedelta(days=long_window)).strftime("%Y-%m-%d")

if args.end_date is not None:
    endDate = args.end_date
else:
    txt = 'Parameter "end date": using default value todays date.'
    logging.info(txt)
    endDate = datetime.now().strftime("%Y-%m-%d")

if valid_date(startDate):
    # evaluate the date parameters given
    dt_startDate = datetime.strptime(startDate, '%Y-%m-%d')
else:
    txt = 'Invalid start date parameter: ' + startDate + '. Ending Script.'
    logging.error(txt)
    quit()

if valid_date(endDate):
    # evaluate the date parameters given
    dt_endDate = datetime.strptime(endDate, '%Y-%m-%d')
else:
    txt = 'invalid end date parameter: ' + endDate + '. Ending Script.'
    logging.error(txt)
    quit()

delta = dt_endDate - dt_startDate
if delta.days < int(long_window):
    txt = "not enough time between dates: " + startDate + ' and ' + endDate + ' for a ' + str(long_window) \
          + ' day window. Ending Script.'
    logging.error(txt)
    quit()

# check verbose for logging
if args.verbose is not None:
    if isinstance(args.verbose, bool):
        verbose = args.verbose
    else:
        txt = 'Parameter "verbose": Using default value False for verbose logging.'
        logging.info(txt)
        verbose = True
else:
    logging.info(txt)
    verbose = True
if verbose:
    log_level = logging.DEBUG
    handler.setLevel(log_level)
    logging.getLogger().setLevel(logging.DEBUG)
    logging.debug('Parameter "verbose": True. Switching on debug logging.')

# SPARK or python
txt = 'Parameter "spark": using default value False. This script will run locally on python.'
if args.spark_flag is not None:
    if isinstance(args.spark_flag, bool):
        spark_flag = args.spark_flag
        logging.info(txt)
    else:
        logging.info(txt)
        spark_flag = False
else:
    logging.info(txt)
    spark_flag = False

# local output mode: ie send trends to a local csv
txt = 'Local file not given. Trends will be passed to db2 api, not saved to local file.'
if args.output_file_name is not None:
    if isinstance(args.output_file_name, str):
        output_file_name = args.output_file_name
        # validate the file name
        p_directory, fname = os.path.split(output_file_name)
        match = re.match('(\w+)\.csv$', fname)
        if match is None:
            txt = 'Invalid output file parameter: ', output_file_name, '.Should be "xxx.csv". Ending Script.'
            logging.error(txt)
            quit()

        # check directory
        out_folder = '../output/'
        if not p_directory:
            txt = 'Pre-pending output directory for output file: ' + out_folder
            logging.info(txt)
            output_file_name = os.path.join(out_folder, output_file_name)
        else:
            # overwrite any other file path
            output_file_name = os.path.join(out_folder, output_file_name)

        if os.path.isfile(output_file_name):
            txt = 'Output file already exists and will be overwritten: ' + output_file_name
            logging.info(txt)
            # check if file is already open
            try:
                myfile = open(output_file_name, "r+")
                myfile.close()
            except IOError:
                txt = 'Output file : ' + output_file_name + ' is already open. Please close this file and re-run. ' \
                                                            'Ending Script.'
                logging.error(txt)
                quit()

    else:
        txt = 'Invalid local file parameter: ' + args.output_file_name + '. Ending Script.'
        logging.error(txt)
        quit()
else:
    txt = 'Local file not given. Trends will be passed to db2 api, not saved to local file.'
    logging.info(txt)
    local = False

# zip code mapping flag
if args.zipcode is not None:
    zipcode = args.zipcode
else:
    zipcode = False

# ssl verification, switch to false for debug only
if args.no_ssl_verification is not None:
    verify = False
else:
    verify = True

# get db2 integration params from config file
if config is not None and db2:
    database = config.get("db2 integration", "database")  # e.g.  # database = SIFS
    table = config.get("db2 integration", "table")  # table = SIFS.COMPLAINT_FEATURES
    # spark may need a db specific driver file e.g.  # driver_jar_path = /home/sifsuser/lib/db2jcc.jar
    driver_jar_path = config.get("db2 integration", "driver_jar_path")

# to do sanity check on these

# check the config
if config is not None:
    mk_confidence = config.getfloat("mann kendall", "mk_confidence")
    min_records = config.getint("trend candidates", "min_records")
    min_per_day = config.getint("trend candidates", "min_per_day")
    pct_change_threshold = config.getfloat("trend candidates", "pct_change_threshold")
    MIN_MONOTONE = config.getfloat("trend candidates", "min_monotone")
    fields_list = config.get("trend candidates", "fields_list")
    fields_list = fields_list.replace('-', '_').split(',')
    # check for filter on products
    products = config.get("trend candidates", "products")
    if products == 'all':
        products = []
    else:
        products = products.split(",")
    model_config = dict(config.items('poisson model'))
    model_name = config.get('poisson model', "model_name")

    # v2 updates
    MAX_RISK = config.getfloat("risk score", "max_risk")  # 85% # 503 max risk score allowable is 85%
    RISK_WEIGHTS = json.loads(config.get("risk score", "risk_weights"))  # rel. weights of factors for scoring trends
    MAX_COUNT = config.getfloat("risk score", "max_count")
    MAX_GRAD = config.getfloat("risk score", "max_grad")
    MIN_RISK_FILTER = config.getfloat("risk score", "min_risk_filter")

    # check validity of risk weights
    if len(RISK_WEIGHTS) != 3:
        # warning default equal weighting used for all weights
        txt = 'Default values are used to give equal weighting to all factors in the risk score calculation.'
        logging.info(txt)
        RISK_WEIGHTS = [1 / 3 for _ in range(3)]
    # normalise weights if required
    if sum(RISK_WEIGHTS) != 1:
        RISK_WEIGHTS = np.asarray(RISK_WEIGHTS) / sum(RISK_WEIGHTS)
else:
    txt = 'Unable to read config file. Ending Script.'
    logging.error(txt)
    quit()

# zip code mapping: map zip code to state
if zipcode:
    # 1 check config file for details of zipcode mapping file
    zipcode_file = config.get("zip code mapping", "zipcode_file")
    # 2 check if file exists
    if not os.path.isfile(zipcode_file):
        txt = 'Unable to find Zip-code to State mapping file ' + zipcode_file + \
              '. Zip code mapping not applied for this run.'
        logging.error(txt)
        zipcode = False
    else:
        df_zip = pd.read_csv(zipcode_file)
        # check file is formatted correctly
        req_fields = ['Zipcode', 'State']
        if any([f for f in req_fields if f not in df_zip.columns]):
            logging.error('missing columns for zip and state')
            zipcode = False
        df_zip['Zipcode'] = df_zip['Zipcode'].astype(str)
        df_zip.set_index('Zipcode', inplace=True)
        zipCodeMap = df_zip.State.to_dict()

# Main processing starts here
# initialise a poisson trend detection model
model = getattr(td_models, model_name)(config=model_config)

# spark section
logging.info('started reading the data')
if spark_flag:  # program run as spark submit job
    spark, df = spark_read_db2(jdbcURL, table, user, passwd, driver_jar_path, tsl)

    # sanity check all the fields in the fields list
    if any([f not in df.columns for f in fields_list]):
        txt = 'Unknown field in fields list. Update the fields list in config.cfg.  Ending Script.'
        logging.error(txt)
        quit()

    func = udf(lambda x: '1' in x or 'TRUE' in x or 'True' in x, BooleanType())
    df = df.withColumn('IS_COMPLAINT_N', func(col('IS_COMPLAINT')))

    # converts the date in US format into a date type:
    func = udf(lambda x: datetime.strptime(x, '%d/%m/%Y'), DateType())
    df = df.withColumn('DATE', func(col('COMPLAINT_DATE')))

    # df.registerTempTable("complaint_features")
    # # dynamically build a query
    # fs = ', '.join(fields_list.copy() + ['DATE', 'PRODUCT', 'ID'])
    # q_sd = '"2017-06-01"'  # needs double quotes
    # # str_query = "select {} from complaint_features where IS_COMPLAINT_N = true and  GEO_REF = 'TX'".format(fs)
    # str_query = "select {} from complaint_features where IS_COMPLAINT_N = true and  GEO_REF = 'TX'" \
    #             " and date(DATE) >= {}".format(fs, q_sd)
    #
    # print(str_query)
    # df_from_query = spark.sql(str_query)
    #
    # # now its available as a df
    # df_from_query.show()
    # # print(type(df_from_query))
    # Pre filtering : apply any global filters ie if we always want to exclude complaints with column X = Y
    if products:
        df = df.filter(col('PRODUCT').isin(products))

    # Cache data: this effectively becomes our starting point for the dataset in spark
    df.cache()

    logging.info('finished reading the data')
    logging.info('# records : {}'.format(df.count()))

    # Create combinations of  fields
    logging.info('find all combinations of fields:' + ', '.join(fields_list))
    field_combinations = [list(subset) for subset in all_subsets(fields_list) if subset]
    # The UI focusses on theme so we only care about combinations that include THEME
    field_combinations = [c for c in field_combinations if 'THEME' in c]

    # start point for the rolling window is the start date + long window .e.g 1st Jan + 30 days
    startDate_rw = dt_startDate + timedelta(days=int(long_window))  # start date of the rolling window
    rolling_window = pd.date_range(start=startDate_rw, end=endDate, freq='1D', name='date range')

    # Now process all candidates and see which actually have trends
    # main loop: loop over 30 day timeframes over a longer window
    complainttrends = []  # trends found will be stored  in here

    keep_cols = ['complaintcount', 'description', 'geo', 'product', 'riskscore', 'theme', 'timeperiod',
                 'trendattributes', 'trendstartdate', 'trendtypeid']

    for day in rolling_window:
        start_date = datetime.strftime(day - timedelta(days=long_window), '%Y-%m-%d')
        start_date_inc_sw = datetime.strftime(day - timedelta(days=long_window + short_window - 1), '%Y-%m-%d')
        end_date = datetime.strftime(day, '%Y-%m-%d')

        logging.info('# complaints before filtering on dates between ' + start_date + ':' + end_date +
                     ' ' + str(df.count()))
        df5 = df.filter(col('DATE') >= start_date).filter(col('DATE') <= end_date)
        logging.info('# complaints after filtering : ' + str(df5.count()))

        # get candidate trends
        candidates = get_candidates_spark(spark.sparkContext, df5, field_combinations, min_records, min_per_day)
        logging.info('found ', str(candidates.count()), ' candidate combinations')
        # spark ends here !
        df_data = df5.toPandas()
        # end of spark
        # Complaint ID is used as the unique id for each complaint in the groupby clauses for counting
        # if each complaint can have multiple combinations then the ID must be made unique
        # other wise the counts will only include each complaint once
        if df_data['ID'].value_counts().max() > 1:
            df_data['ID'] = str(df_data['ID']) + df_data.index.astype(str)

        df_data_filtered = df_data[(df_data.date >= start_date_inc_sw) & (df_data.date <= end_date)]
        if (df_data_filtered is None) | (len(df_data_filtered) == 0):
            logging.info('No complaints found')
            continue
        else:
            logging.info('# complaints after filtering: ' + str(len(df_data_filtered)))


# python
else:  # program run as python xxx.py version
    # read data from hdfs, db2 or local file system
    if args.input_file_name is not None and 'hdfs' in input_file_name:
        df_data = hdfs_read(input_file_name).toPandas()  # probably redundant method to read input from hds
        df_data.drop(0, inplace=True)
    elif db2:
        _, df_spark = spark_read_db2(jdbcURL, table, user, passwd, driver_jar_path, tsl)  # don't need ref to spark session
        df_data = df_spark.toPandas()
    else:  # local file
        df_data = pd.read_csv(input_file_name, low_memory=False, quoting=0, escapechar='\\')
    logging.info('finished reading the data')
    # replace spaces in column headers
    schemaString_dict = {s: s.replace(' ', '_').replace('-', '_') for s in df_data.columns}
    df_data.rename(columns=schemaString_dict, inplace=True)

    # sanity check all the fields in the fields list
    if any([f not in df_data.columns for f in fields_list]):
        logging.error('Unknown field in fields list. Update the fields list in config.cfg.  Ending Script.')
        quit()

    # pre-processing in pandas
    # filter on complaints
    df_data = df_data[df_data['IS_COMPLAINT']== '1']  # for spark read everything is initially read as a string

    # create a new date column with a datetype
    df_data['date'] = df_data['COMPLAINT_DATE']  # previously COMPLAINT_DATE was a string, needed to be parsed to date

    # make complaint ID unique:
    # Complaint ID is used as the unique id for each complaint in the groupby clauses for counting
    # if each complaint can have multiple combinations then the ID must be made unique
    # otherwise the counts will only include each complaint once
    if df_data['ID'].value_counts().max() > 1:
        df_data['ID'] = str(df_data['ID']) + df_data.index.astype(str)

    # globally filter the dataset
    logging.info('# records before filtering:' + str(len(df_data)))
    if products:
        df_data = df_data[df_data.PRODUCT.isin(products)]
        logging.info('# records after filtering on products : ' + str(len(df_data)))

    # turn the age field into categorical
    if 'CUSTOMER_AGE' in fields_list:
        df_data.CUSTOMER_AGE = pd.to_numeric(df_data.CUSTOMER_AGE)
        df_data.loc[df_data.CUSTOMER_AGE < 0, 'CUSTOMER_AGE'] = 0
        df_data.loc[df_data.CUSTOMER_AGE > 90, 'CUSTOMER_AGE'] = 89
        df_data['CUSTOMER_AGE_RAW'] = pd.to_numeric(df_data.CUSTOMER_AGE)  # copy unbinned age
        labels = ['0-30', '30-40', '40-50', '50-60', '60-100']
        bins = [0, 30, 40, 50, 60, 120]
        df_data['CUSTOMER_AGE'] = pd.cut(df_data.CUSTOMER_AGE, bins=bins, right=False, labels=labels)

    if ('STATE' in fields_list) & zipcode:
        df_data['GEO_REF'] = df_data['GEO_REF'].fillna(0)
        df_data['GEO_REF'] = df_data['GEO_REF'].map(check_num)
        df_data['GEO_REF'] = df_data['GEO_REF'].astype(int).astype(str)
        df_data['STATE'] = df_data['GEO_REF'].map(zipCodeMap)

    # process reference fields
    ref_fields = ['PRODUCT_REF', 'SUB_PRODUCT_REF', 'GEO_REF']
    for rf in ref_fields:
        replace_field = rf.replace('_REF', '')
        # check that both original and ref arent in list
        if (rf in fields_list) & (replace_field in fields_list):
            txt = 'Trend fields include original and reference: ' + replace_field + ' , ' + rf + \
                  '. Change config file to select only one. Ending Script.'
            logging.error(txt)
            quit()

        if rf in fields_list:
            df_data[replace_field] = df_data[rf]  # overwrite the field eg Product_REF -> Product
            loc = fields_list.index(rf)
            fields_list[loc] = replace_field

    # Create combinations of  fields
    logging.info('find all combinations of fields:' + ', '.join(fields_list))
    field_combinations = [list(subset) for subset in all_subsets(fields_list) if subset]
    # The UI focusses on theme so we only care about combinations that include THEME
    field_combinations = [c for c in field_combinations if 'THEME' in c]

    # start point for the rolling window is the start date + long window .e.g 1st Jan + 30 days
    startDate_rw = dt_startDate + timedelta(days=int(long_window))  # start date of the rolling window
    rolling_window = pd.date_range(start=startDate_rw, end=endDate, freq='1D', name='date range')

    # Now process all candidates and see which actually have trends
    # main loop: loop over 30 day timeframes over a longer window
    complainttrends = []  # trends found will be stored  in here

    keep_cols = ['complaintcount', 'description', 'geo', 'product', 'riskscore', 'theme', 'timeperiod',
                 'trendattributes', 'trendstartdate', 'trendtypeid']

    for day in rolling_window:
        start_date = datetime.strftime(day - timedelta(days=long_window), '%Y-%m-%d')
        start_date_inc_sw = datetime.strftime(day - timedelta(days=long_window + short_window - 1), '%Y-%m-%d')
        end_date = datetime.strftime(day, '%Y-%m-%d')

        if df_data is not None:
            logging.info('# complaints before filtering on dates between ' + start_date + ' : ' + end_date +
                         ' ' + str(len(df_data)))

        df_data_filtered = df_data[(df_data.date >= start_date_inc_sw) & (df_data.date <= end_date)]
        if (df_data_filtered is None) | (len(df_data_filtered) == 0):
            logging.info('No complaints found')
            continue
        else:
            logging.info('# complaints after filtering: ' + str(len(df_data_filtered)))

        # get a fixed index over every day in start date -> end date
        fixed_index = pd.date_range(start=start_date_inc_sw, end=end_date, freq='1D', name='date')

        # process all candidates and see which have increasing trends
        df_c = pd.DataFrame()
        df_c['selection_criteria'] = get_candidates(df_data_filtered, field_combinations, min_records, min_per_day)
        logging.info('# candidate trends found: {}'.format(len(df_c)))
        df_c['date_range'] = start_date + ':' + end_date

        df_c['trend'], df_c['datecounts'] = zip(*df_c['selection_criteria'].apply(create_timeline,
                                                                                  args=(df_data_filtered, fixed_index,
                                                                                        short_window,
                                                                                        long_window, MIN_MONOTONE,
                                                                                        pct_change_threshold)))

        # filter out all rows where trend is empty
        df_c = df_c[[bool(i) for i in df_c.trend]]

        # filter out duplicate rows
        # The list of combinations is now converted to a dataframe again ( python ok here as its low volume
        # ie max hundreds for any one day)
        # Its possible that a trend can be duplicated due to the method that trends are discovered
        # ie select all trends where issue = mortage  and
        # select all trends where issue = mortgage and state = texas will find the same trend
        # if all mortgage issues happened to have occured in texas
        # the code below removes duplicates and keeps the lowest level description of any duplicate trend
        # add number of selection attributes for each detected trend
        df_c['num_attributes'] = df_c.selection_criteria.map(len)
        df_c['max_attributes'] = 0
        unique_trends = list(set([tuple(x[0]) for x in df_c.trend]))  # get all trends
        df_c['trend_idx'] = df_c['trend'].apply(lambda z: unique_trends.index(tuple(z[0])))
        max_map = df_c[['trend_idx', 'num_attributes']].groupby(['trend_idx']).max().to_dict()['num_attributes']
        df_c['max_attributes'] = df_c['trend_idx'].map(max_map)
        df_c = df_c[df_c['num_attributes'] == df_c['max_attributes']]

        # analyze trends and detect trends
        df_c['summary'] = df_c['trend'].apply(analyze_trend_new, args=(fixed_index, short_window, model, mk_confidence,
                                                                       float(model_config['poisson_threshold'])))
        df_c = df_c.apply(detect_trend_new, axis=1, args=(short_window, MAX_RISK, MAX_COUNT, RISK_WEIGHTS, MAX_GRAD))

        # apply trend filter
        if 'trend_detected' in df_c.columns:
            df_c = df_c[df_c.trend_detected]
        else:
            logging.info('no trends detected in the candidate trends.')
            continue

        if len(df_c) == 0:
            logging.info('no trends detected in the candidate trends.')
            continue

        # apply the min_risk_score filter: drop any trends with risk score below a minimum amount
        df_c["riskscore"] = df_c['summary'].map(lambda x: x.get('riskscore', 0.0))
        drop_trends = sum(df_c["riskscore"] < MIN_RISK_FILTER)

        if drop_trends:
            df_c = df_c[df_c["riskscore"] >= MIN_RISK_FILTER]
        if len(df_c) == 0:
            logging.info('No trends above minimum risk score threshold.')
            continue

        # fill in details for each trend
        df_c['timeperiod'] = long_window
        df_c['description'] = df_c['selection_criteria'].map(
            lambda x: "&".join([k.replace('&', '_') + ':' + str(v).replace('&', '_') for k, v in x.items()]))
        df_c['trendattributes'] = df_c['selection_criteria'].map(map_selections)
        df_c['product'] = df_c['selection_criteria'].map(lambda x: x.get('PRODUCT', 'all'))
        df_c['geo'] = df_c['selection_criteria'].map(lambda x: x.get('GEO', 'all'))
        df_c['theme'] = df_c['selection_criteria'].map(lambda x: x.get('THEME', "all"))
        df_c["trendtypeid"] = df_c['summary'].map(lambda x: x.get('trendtypeid', 0))
        df_c = df_c.apply(create_complaint_count, axis=1)

        # append results to output structure
        trends = df_c[keep_cols].to_dict(orient='records')
        complainttrends.extend(trends)
        # logging
        logging.info('found ' + str(len(trends)) + ' new trend(s)')
        logging.info('# Trends dropped as risk score below minimum (' + str(MIN_RISK_FILTER) + '): ' + str(drop_trends))
        logging.info('Total trends found so far: ' + str(len(complainttrends)))
        for i, t in enumerate(trends):
            logging.info('new trend {:02d}:'.format(i + 1))
            logging.info(json.dumps(t))

        # write results to database using API
        rest_api = True
        if args.output_file_name is None:
            if rest_api:
                # send the results to db2 database using the api call
                logging.info('upload trends to db2 using api')
                # send array of  trends to be persisted in db2
                # params = dict()  # dict for the params
                # params['complainttrends'] = trends  # param1 = detected trends send as a list
                # response = requests.post(api_url, json=params, verify=verify)
                response = post_request(trends, api_url, verify)
                logging.info("response status code " + str(response.status_code))
                logging.info("response" + response.text)
                if response.status_code != 200:
                    for p in complainttrends:
                        logging.error('db update error')
                        logging.error('database update failed for: ' + json.dumps(p))

    # write results to local csv
    if len(complainttrends):
        if args.output_file_name is not None:
            logging.info('saving results to local file')
            df = pd.DataFrame(complainttrends)
            df.to_csv(output_file_name, index=False)

    # final summary
    logging.info('Found ' + str(len(complainttrends)) + ' trends')
    logging.info('All done. Ending script.')

# reset logging level to avoid errors after 'tear-down'
log_level = logging.INFO
handler.setLevel(log_level)
logging.basicConfig(level=log_level, format=log_format, handlers=[handler])
