# common functions used for trend detection
from datetime import datetime, timedelta
import os
import re
import json
import configparser
import math
import numpy as np
import pandas as pd
from itertools import chain, combinations
import requests
import ibm_db
import ibm_db_dbi
# Trend Detection using the gnip models
import gnip_trend_detection.models as td_models
import gnip_trend_detection.analysis as td_analysis
# spark
from pyspark import SparkConf, SparkContext
from pyspark.sql import SparkSession
from pyspark.sql.types import StructField, StructType, StringType, LongType, DateType
from pyspark.sql import SQLContext
from pyspark.sql.functions import col, udf
from pyspark.sql.functions import datediff, months_between, to_date
from pyspark.sql.types import *
from pyspark.sql import Row
import base64

def valid_date(datestring):
    try:
        datetime.strptime(datestring, '%Y-%m-%d')
        return True
    except ValueError:
        return False


def hdfs_read(input_file):
    """
    read a file from hdfs and return a spark dataframe
    :param input_file:
    :return:
    """
    app_name = "Complaint Trend Detection"
    # Configure Spark
    conf = SparkConf().setAppName(app_name)
    conf = conf.setMaster("local[*]")  # to do pull out the spark config
    sc = SparkContext(conf=conf)
    sqlContext = SQLContext(sc)

    # read the input file
    dataRDD = sc.textFile(input_file)
    temp_rdd = dataRDD.mapPartitions(lambda x: csv.reader(x, delimiter=","))

    # Extract the header line and use this to create a schema
    header = [c.replace(' ', '_').replace('-', '_') for c in dataRDD.first().split(',')]

    # Assuming that all the columns are string, create a new StructField for each column
    fields = [StructField(field_name, StringType(), True) for field_name in header]
    schema = StructType(fields)  # construct our schema,

    # The Schema is passed in as a parameter into the createDataFrame() function
    df1 = sqlContext.createDataFrame(temp_rdd, schema)
    return df1


def spark_read_db2(jdbcurl, tbl, usr, pwd, driver_jar_path, tsl):
    """
    :param jdbcurl:
    :param tbl:
    :param usr:
    :param pwd:
    :param driver_jar_path:
    :param tsl:
    :return:
    """
    spark = SparkSession.builder \
        .appName("spark_test") \
        .config("spark.driver.extraClassPath", driver_jar_path) \
        .getOrCreate()

    df = spark.read \
        .format("jdbc") \
        .option("url", jdbcurl) \
        .option("sslTrustStoreLocation", tsl)\
        .option("dbtable", tbl) \
        .option("user", usr) \
        .option("password", pwd) \
        .load()

    return spark, df


def calc_percent_change(count_series, end_idx, previous_idx):
    """
    # e.g find the % increase over the last 7 days
    7 days ago 3 complaints per day, today 9 complaints a day
    % increase = 9-3 / 3 = 200 % increase
    original value is 3 , a 100% increase would be 6, a 200% increase would be 9

    Parameters:
    count_series is a pandas series
    end_idx is the index of the last value in pandas indexing ie -1
    previous_idx is the index of the before value ie the value 7 days before the end of the series = -7
    """
    # check that there is a value 30 and 7 days ago ie that there are enough days in the timeseries
    if (previous_idx * -1) > len(count_series):
        return 0

    count_end = max(count_series[end_idx], 1)  # use max(x,1) because values lower than 1 give crazy values
    count_previous = max(count_series[previous_idx], 1)

    sign = '+'
    if count_previous == 0:
        pct_change = 0
    else:
        pct_change = np.round(100 * (count_end - count_previous) / count_previous, 2)
        if count_end < count_previous:
            sign = ''

    return sign + str(pct_change) + '%'


def f_val_in_dict(drop_dict, **kwargs):
    """
    filter a dataframe dynamically: column:values must be in dict
    :param drop_dict:
    :return: boolean value
    """
    return kwargs in drop_dict


def all_subsets(ss):
    """
    generate all combinations of fields ( this is  generator)
    :param ss:
    :return:
    """
    return chain(*map(lambda x: combinations(ss, x), range(0, len(ss) + 1)))


def get_candidates_spark(sc, data_df, combns, minrecs, minperday):
    """
    candidate = a valid combination of fields and values that could represent a trend
    Spark function to find combinations of data

    :param sc:
    :param data_df:
    :param combns:  all combinations of the fields
    :param minrecs:
    :param minperday:
    :return:
    """
    allfields = set([i for row in combns for i in row])  # get all the unique fields
    counts = []
    maxes = []

    for comb in combns:
        # filter 1: there must be at least <<min records>> total records in this timeline
        select_fields = tuple(comb) + ('ID',)
        groupby_fields = tuple(comb)
        counts.append(data_df.selectExpr(*select_fields).
                      groupBy(*groupby_fields).
                      count().
                      where(col('count') > minrecs).
                      collect())

        # filter 2: there must be at least << min per day>> complaints on one day in this timeline
        select_fields = tuple(comb) + ('ID', 'Date')
        groupby_fields = tuple(comb) + ('Date',)
        maxes.append(data_df.selectExpr(*select_fields).
                     groupBy(*groupby_fields).
                     count().
                     groupBy(*tuple(comb)).
                     agg({'count': 'max'}).
                     where(col('max(count)') > minperday).
                     collect())

    rdd_counts1 = sc.parallelize([i for row in counts for i in row])  # flatten
    rdd_maxes1 = sc.parallelize([i for row in maxes for i in row])  # flatten

    # create hashable keys
    rdd_counts2 = rdd_counts1.map(lambda x: tuple((k, x[k]) for k in allfields if k in x))
    rdd_maxes2 = rdd_maxes1.map(lambda x: tuple((k, x[k]) for k in allfields if k in x))
    rdd_maxes2.take(10)

    # keep the candidates that meet both
    rdd_candidates = rdd_counts2.intersection(rdd_maxes2)
    rdd_candidates2 = rdd_candidates.map(lambda x: {k: v for k, v in x})
    return rdd_candidates2


def get_candidates(data_df, combs, min_recs, min_pd):
    """
    candidate = a valid combination of fields and values that could represent a trend
    a candidate is a set of dimensions that might hold a timeline
    eg issue = overcharged, product = personal loan, geo = NY

    :param data_df:
    :param combs: all combinations of the fields
    :param min_recs minimum number of records before it is considered a valid candidate
    :param min_pd the smallest max number of records per day in the series before it is considered a valid candidate
    :return:

    """
    cands = []
    for field_list in combs:
        # filter 1) only bother looking for timelines where there are at least 10 records
        df_count = data_df[field_list + ['ID']].groupby(field_list).count()  # Complaint ID used to count
        df_combinations = df_count[df_count['ID'] > min_recs].reset_index()[field_list]

        # filter 2 count by day and get the max
        # eg df_data[['Product', 'Date','Complaint ID']].groupby(['Product', 'Date']).count().groupby(['Product']).max()
        df_temp = data_df[field_list + ['ID', 'date']].groupby(field_list + ['date']).count().groupby(
            field_list).max()

        # find the ocmbinations that can be dropped due to filter 2
        dict_drop_combinations = df_temp[df_temp['ID'] < min_pd].reset_index()
        if dict_drop_combinations is not None:
            if all([f in dict_drop_combinations.columns for f in field_list]):
                dict_drop_combinations = dict_drop_combinations[field_list].to_dict(orient='records')

                df_combinations = df_combinations[
                    df_combinations[field_list].apply(lambda x: not (f_val_in_dict(dict_drop_combinations, **x)),
                                                      axis=1)]

                cands.extend(df_combinations.to_dict(orient='records'))
            else:
                continue
        else:
            continue
    return cands


def analyze_trend_new(counts, date_index, trend_win_length, model_, mk_conf, poisson_threshold):
    """
    looks at a long window of counts and see if a trend is occurring in the trend window
    or last 7 days

    :param counts:      a list of raw counts over the long window
    :param date_index:
    :param trend_win_length: ie short window of 7 days
    :param model_:
    :param mk_conf mann-kendall confidence threshold
    :param poisson_threshold:
    :return: a dictionary of summary stats per trend
    """
    mk_dict = {'increasing': 1, 'decreasing': 2, 'no trend': 3}

    ts = np.asarray(counts[0])

    count_in_trend_window = ts[-trend_win_length:].sum()
    mean_count = ts.mean()
    mean_count_in_trend_window = ts[-trend_win_length:].mean()
    mean_ratio = 0
    if mean_count:
        mean_ratio = mean_count_in_trend_window / mean_count  # ratio of mean in trend window to mean of whole period

    # poisson model is used to identify exceptional high or low counts
    spd = 86400  # number of seconds in a day
    pm_values = [[datetime.strftime(d, '%Y-%m-%d'), s, c]
                 for (d, s, c) in zip(date_index, [spd for _ in range(len(ts))], counts[0])]

    pm_scores = td_analysis.analyze(pm_values, model_)[-trend_win_length:]
    poisson_scores_max = max([i[2] for i in pm_scores])
    poisson_above_thresh_count_inc = sum([1 for i in pm_scores if (i[2] > poisson_threshold) & (i[1] > mean_count)])
    poisson_above_thresh_count_dec = sum([1 for i in pm_scores if (i[2] > poisson_threshold) & (i[1] < mean_count)])

    moving_average = counts[1][-trend_win_length:]
    mann_kendall_score = td_models.mk_test(moving_average, mk_conf)

    # calculate the monotonicity score depending on trend type: 'decreasing':2, 'increasing':1, 'no trend':3
    monotone_score = 0
    if mann_kendall_score[0] == 'increasing':
        monotone_score = sum([x < y for x, y in zip(ts[-trend_win_length:], ts[-(trend_win_length - 1):])]) \
                         / (trend_win_length - 1)
    if mann_kendall_score[0] == 'decreasing':
        monotone_score = sum([x > y for x, y in zip(ts[-trend_win_length:], ts[-(trend_win_length - 1):])]) \
                         / (trend_win_length - 1)

    # output results to a dictionary
    summary = {
        'count': round(ts.sum(), 2),
        'count in trend window': round(count_in_trend_window, 2),
        'mean count': float("%.2f" % round(mean_count, 2)),
        'mean count in trend window': float("%.2f" % round(mean_count_in_trend_window, 2)),
        'mean ratio': round(mean_ratio, 2),
        # increase in % over the last 7 days using the smooth counts
        'trendinglast7days': calc_percent_change(counts[1], -1, -7),
        # increase in % over the last 30 days using the smooth counts
        'trendinglast30days': calc_percent_change(counts[1], -1, -30),
        'trendtypeid': mk_dict[mann_kendall_score[0]],  # 'decreasing':2, 'increasing':1, 'no trend':3
        'mk trend': mann_kendall_score[1],
        'mk score p-value': round(mann_kendall_score[2], 3),
        'poisson_scores_max': poisson_scores_max,
        'poisson_above_thresh_count_inc': poisson_above_thresh_count_inc,
        'poisson_above_thresh_count_dec': poisson_above_thresh_count_dec,
        'trend_detected': False,  # set default value
        'riskscore': 0.0,  # set default value
        'monotonicity': monotone_score
    }
    return summary


def exp_smooth(points, factor=0.8):
    """
    exponential smoothing
    :param points:
    :param factor:
    :return:
    """
    smoothed_points = []
    for point in points:
        if smoothed_points:
            previous = smoothed_points[-1]
            smoothed_points.append(previous * factor + point * (1 - factor))
        else:
            smoothed_points.append(point)
    return smoothed_points


def detect_trend_new(row, win, risk_max, count_max, risk_wts, grad_max):
    risk_max = float(risk_max)

    # get trend details
    summary = row['summary']  # dictionary of summary stats for a trend
    counts = row['trend'][0]  # 0 = list of raw counts, 1 = list of smooth counts

    # trend start and end date
    s_date, e_date = row['date_range'].split(':')
    dates = pd.date_range(start=s_date, end=e_date, freq='1D', name='Date')

    # set defaults
    trend_detected = False
    trend_start_index = None
    trend_end_index = None
    trend_start_value = 0
    trend_end_value = 0
    row["trendstartdate"] = None

    # rising trends
    if (summary['mk trend']) & \
            (summary['trendtypeid'] == 1) & \
            bool(summary['poisson_above_thresh_count_inc']):

        # trend start date: 1st date in short window greater than long term average
        for i, v in enumerate(counts[-win:]):
            if v > np.mean(counts):  # average over long window
                trend_start_index, trend_start_value = -win + i, v
                trend_start_date = dates[trend_start_index]
                row["trendstartdate"] = datetime.strftime(trend_start_date, '%Y-%m-%d')
                break

        # trend end date: latest date in short window greater than the value at the start date
        for i, v in enumerate(counts[-1:-(win + 1):-1]):
            if v > trend_start_value:
                trend_end_index, trend_end_value = -(i + 1), v
                trend_end_date = dates[trend_end_index]
                row["trendenddate"] = datetime.strftime(trend_end_date, '%Y-%m-%d')
                break

    elif (summary['mk trend']) & \
            (summary['trendtypeid'] == 2) & \
            bool(summary['poisson_above_thresh_count_dec']):

        # trend start date: 1st date in short window greater than long term average
        for i, v in enumerate(counts[-win:]):
            if v < np.mean(counts):  # average over long window
                trend_start_index, trend_start_value = -win + i, v
                trend_start_date = dates[trend_start_index]
                row["trendstartdate"] = datetime.strftime(trend_start_date, '%Y-%m-%d')
                break

        # trend end date: latest date in short window greater than the value at the start date
        for i, v in enumerate(counts[-1:-(win + 1):-1]):
            if v < trend_start_value:  # average over long window
                trend_end_index, trend_end_value = -(i + 1), v
                trend_end_date = dates[trend_end_index]
                row["trendenddate"] = datetime.strftime(trend_end_date, '%Y-%m-%d')
                break

    # if trend start and end date have been detected
    if bool(trend_start_index) & bool(trend_end_index):

        grad = abs(trend_end_value - trend_start_value) / (trend_end_index - trend_start_index)
        if grad >= 1.0:  # ie trend shows an increase or decrease of at least one complaint per day
            # find number of complaints due to trend:  count in trend window minus the expected (average)
            # for a falling trend this would be negative, hence the abs
            summary['count_in_trend'] = abs(sum(counts[trend_start_index:trend_end_index]) -
                                            (trend_end_index - trend_start_index) * summary['mean count'])
            if summary['count_in_trend'] > 1.0: # must be at least one difference from expected mean count
                trend_detected = True
                summary['gradient'] = grad

                # calculate risk score
                summary['riskscore'] = calc_riskscore(summary, risk_max, count_max, risk_wts, grad_max)

    row['summary'] = summary
    row['trend_detected'] = trend_detected

    return row


# fast version to check times
def create_timeline(select_expr, data_df, date_index, sw, lw, monotone_min, pct_change_thresh):
    """
    create  a new data frame with Date as index and record count as the only column
    :return:
    :return:
    :param  select_expr:
    :param  data_df:
    :param  date_index: all the days in the long window in date format as an index
    :param sw short window length , normally 7 days
    :param lw long window, eg 30 days
    :param monotone_min
    :param pct_change_thresh  minimum change threshold for any trend
    :return: a list of complaint counts per day for this select expression,

    """
    pass_min_threshold_checks = False
    datecounts = []

    # filter data set on expression
    expr = ' & '.join([str(k) + ' == "' + str(v) + '"' for k, v in select_expr.items()])
    df_temp = data_df.query(expr)

    # groupby date and count per day
    cols = [k for k in select_expr.keys()]
    values = tuple(select_expr.values())
    df_timeline = df_temp.set_index(cols).sort_index()
    df_timeline = df_timeline.loc[values, ['date', 'ID']].groupby('date').count()

    # re-index to have a consistent start and end date for all timelines
    df_timeline = df_timeline.reindex(index=date_index, fill_value=0)

    # rename to get the counts columns
    df_timeline.rename(columns={'ID': 'actualcount'}, inplace=True)
    complaint_counts = list(df_timeline['actualcount'])[-lw:]

    # check for generally upward or downward trend
    counts_short_window = list(df_timeline['actualcount'][-sw:])

    # apply the minumum change over the long term average threshold
    # using config param: pct_change_threshold
    mean_count = np.asarray(complaint_counts)[-lw:].mean()
    mean_count_in_trend_window = np.asarray(complaint_counts)[-sw:].mean()

    if abs((mean_count_in_trend_window - mean_count) / mean_count) >= pct_change_thresh:
        # number of days in trend that should increase monotonically
        # eg ( 7 days - 1 ) * 0.5 = 3. so must be 3 days with greater count than preceeding day
        min_monotone = int(sw - 1) * monotone_min

        # in the short window at least half of the counts must be greater than the predecessor
        if (sum([x < y for x, y in zip(counts_short_window, counts_short_window[1:])]) >= min_monotone) & \
                (sum([y - x for x, y in zip(counts_short_window, counts_short_window[1:])]) > 0):
            # 'its a relaxed monotonic increase'
            pass_min_threshold_checks = True

        # in the short window at least half of the counts must be less than the predecessor
        if (sum([x > y for x, y in zip(counts_short_window, counts_short_window[1:])]) >= min_monotone) & \
                (sum([y - x for x, y in zip(counts_short_window, counts_short_window[1:])]) < 0):
            # ( 'its a relaxed monotonic decrease')
            pass_min_threshold_checks = True

    # create the datecounts structure for the UI
    if pass_min_threshold_checks:
        # get rolling average, lagged by win_length days ( so get na values for first 6 days if win_length = 7)
        df_timeline['smoothcount'] = np.round(df_timeline['actualcount'].rolling(window=sw).mean(), 2)
        df_timeline.fillna(0, inplace=True)
        df_timeline.reset_index(inplace=True)
        df_timeline['date'] = df_timeline['date'].map(lambda x: x.strftime('%Y-%m-%d'))
        datecounts = df_timeline.iloc[-lw:].to_dict(orient='records')
        # create a list: 1st element = list of raw counts, 2nd element = list of smooth counts
        daily_counts = [complaint_counts, list(df_timeline['smoothcount'])[-lw:]]
    else:
        daily_counts = []
    return daily_counts, datecounts


def od_get_trends(raw_counts, smooth_counts, date_index, sw, lw, monotone_min, model_,
                  mk_conf, pois_thresh, pct_change_thresh, risk_max, count_max, risk_wts, grad_max,
                  minrecords, minperday):
    """
    for the on-demand trend detection, this function does all the trend trend detection work.
    input is a timeline ( ie dates and counts ) for a complaint
    plus other timeline paramaters

    :param raw_counts: list of counts of complaints per day
    :param smooth_counts:
    :param date_index:
    :param sw:
    :param lw: long window, eg 30 days
    :param monotone_min:
    :param model_:
    :param mk_conf:
    :param pois_thresh:
    :param pct_change_thresh: minimum change threshold for any trend
    :param risk_max:
    :param count_max:
    :param risk_wts:
    :param grad_max:
    :param minrecords:
    :param minperday:
    :return: a dictionary of trend statistics
    """

    SECONDS_PER_DAY = 86400  # number of seconds in a day
    pass_min_threshold_checks = False
    mk_dict = {'increasing': 1, 'decreasing': 2, 'no trend': 3}
    risk_max = float(risk_max)

    counts = np.asarray(raw_counts)
    counts_sw = np.asarray(raw_counts)[-sw:]
    mean_count_lw = np.asarray(raw_counts)[-lw:].mean()
    mean_count_sw = np.asarray(raw_counts)[-sw:].mean()

    # apply the minimum (config param: pct_change_threshold) change over the long term average threshold
    if mean_count_lw and abs((mean_count_sw - mean_count_lw) / mean_count_lw) >= pct_change_thresh:
        # number of days in trend that should increase monotonically
        # eg ( 7 days - 1 ) * 0.5 = 3. so must be 3 days with greater count than preceeding day
        min_monotone = int(sw - 1) * monotone_min

        # in the short window at least half of the counts must be greater than the predecessor
        if (sum([x < y for x, y in zip(counts_sw, counts_sw[1:])]) >= min_monotone) & \
                (sum([y - x for x, y in zip(counts_sw, counts_sw[1:])]) > 0):
            # 'its a relaxed monotonic increase'
            pass_min_threshold_checks = True

        # in the short window at least half of the counts must be less than the predecessor
        if (sum([x > y for x, y in zip(counts_sw, counts_sw[1:])]) >= min_monotone) & \
                (sum([y - x for x, y in zip(counts_sw, counts_sw[1:])]) < 0):
            # ( 'its a relaxed monotonic decrease')
            pass_min_threshold_checks = True

    if max(counts) < minperday and sum(counts) < minrecords:
        pass_min_threshold_checks = False

    if pass_min_threshold_checks:
        # ratio of mean in trend window to mean of whole period
        mean_ratio = mean_count_sw / mean_count_lw

        # poisson model is used to identify exceptional high or low counts
        pm_values = [[datetime.strftime(d, '%Y-%m-%d'), s, c]
                     for (d, s, c) in zip(date_index, [SECONDS_PER_DAY for _ in raw_counts], raw_counts)]

        pm_scores = td_analysis.analyze(pm_values, model_)[-sw:]
        poisson_scores_max = max([i[2] for i in pm_scores])
        pois_above_thresh_count_inc = sum([1 for i in pm_scores if (i[2] > pois_thresh) & (i[1] > mean_count_lw)])
        pois_above_thresh_count_dec = sum([1 for i in pm_scores if (i[2] > pois_thresh) & (i[1] < mean_count_lw)])

        # calculate mann-kendall score on the smoothed counts in the short window
        moving_average = smooth_counts[-sw:]
        mann_kendall_score = td_models.mk_test(moving_average, mk_conf)

        # calculate the monotonicity score depending on trend type: 'decreasing':2, 'increasing':1, 'no trend':3
        monotone_score = 0
        if mann_kendall_score[0] == 'increasing':
            monotone_score = sum([x < y for x, y in zip(raw_counts[-sw:], raw_counts[-(sw - 1):])]) / (sw - 1)
        if mann_kendall_score[0] == 'decreasing':
            monotone_score = sum([x > y for x, y in zip(raw_counts[-sw:], raw_counts[-(sw - 1):])]) / (sw - 1)

        # output results to a dictionary
        summary = {
            'key_date': date_index[-1].strftime("%Y-%m-%d"),
            'count': round(counts.sum(), 2),
            'count in trend window': round(counts_sw.sum(), 2),
            'mean count': float("%.2f" % round(mean_count_lw, 2)),
            'mean count in trend window': float("%.2f" % round(mean_count_sw, 2)),
            'mean ratio': round(mean_ratio, 2),
            'trendtypeid': mk_dict[mann_kendall_score[0]],  # 'decreasing':2, 'increasing':1, 'no trend':3
            'mk trend': mann_kendall_score[1],
            'mk score p-value': round(mann_kendall_score[2], 3),
            'poisson_scores_max': poisson_scores_max,
            'poisson_above_thresh_count_inc': pois_above_thresh_count_inc,
            'poisson_above_thresh_count_dec': pois_above_thresh_count_dec,
            'monotonicity': monotone_score
        }

        # set defaults
        trend_start_index = None
        trend_end_index = None
        trend_start_value = 0
        trend_end_value = 0

        # rising trends
        if (summary['mk trend']) & \
                (summary['trendtypeid'] == 1) & \
                bool(summary['poisson_above_thresh_count_inc']):

            # trend start date: 1st date in short window greater than long term average
            for i, v in enumerate(counts[-sw:]):
                if v > np.mean(counts):  # average over long window
                    trend_start_index, trend_start_value = -sw + i, v
                    trend_start_date = date_index[trend_start_index]
                    trendstartdate = datetime.strftime(trend_start_date, '%Y-%m-%d')
                    break

            # trend end date: latest date in short window greater than the value at the start date
            for i, v in enumerate(counts[-1:-(sw + 1):-1]):
                if v > trend_start_value:
                    trend_end_index, trend_end_value = -(i + 1), v
                    trend_end_date = date_index[trend_end_index]
                    trendenddate = datetime.strftime(trend_end_date, '%Y-%m-%d')
                    break

        elif (summary['mk trend']) & \
                (summary['trendtypeid'] == 2) & \
                bool(summary['poisson_above_thresh_count_dec']):

            # trend start date: 1st date in short window greater than long term average
            for i, v in enumerate(counts[-sw:]):
                if v < np.mean(counts):  # average over long window
                    trend_start_index, trend_start_value = -sw + i, v
                    trend_start_date = date_index[trend_start_index]
                    trendstartdate = datetime.strftime(trend_start_date, '%Y-%m-%d')
                    break

            # trend end date: latest date in short window greater than the value at the start date
            for i, v in enumerate(counts[-1:-(sw + 1):-1]):
                if v < trend_start_value:  # average over long window
                    trend_end_index, trend_end_value = -(i + 1), v
                    trend_end_date = date_index[trend_end_index]
                    trendenddate = datetime.strftime(trend_end_date, '%Y-%m-%d')
                    break

        # if trend start and end date have been detected
        if bool(trend_start_index) & bool(trend_end_index):
            grad = abs(trend_end_value - trend_start_value) / (trend_end_index - trend_start_index)
            if grad >= 1.0:  # ie trend shows an increase or decrease of at least one complaint per day
                summary['gradient'] = grad
                summary['trendstartdate'] = trendstartdate
                summary['trendenddate'] = trendenddate
                # find number of complaints due to trend:  count in trend window minus the expected (average)
                # for a falling trend this would be negative, hence the abs
                summary['count_in_trend'] = abs(sum(counts[trend_start_index:trend_end_index]) -
                                                (trend_end_index - trend_start_index) * summary['mean count'])
                # calculate risk score
                summary['riskscore'] = calc_riskscore(summary, risk_max, count_max, risk_wts, grad_max)
                # increase in % over the last 7 days using the smooth counts
                summary['trendinglast7days'] = calc_percent_change(smooth_counts, -1, -7)
                # increase in % over the last 30 days using the smooth counts
                summary['trendinglast30days'] = calc_percent_change(smooth_counts, -1, -30)
                return summary


def get_min_stats(raw_counts, date_index, sw, lw):
    """
    in the case that no trend has been found, this func creates a minimal set of summary statistics for the timeline

    :param raw_counts:
    :param date_index:
    :param sw:
    :param lw:
    :return: summary dictionary structure
    """
    counts = np.asarray(raw_counts)
    counts_sw = np.asarray(raw_counts)[-sw:]
    mean_count_lw = np.asarray(raw_counts)[-lw:].mean()
    mean_count_sw = np.asarray(raw_counts)[-sw:].mean()
    if mean_count_lw > 0:
        # ratio of mean in trend window to mean of whole period
        mean_ratio = mean_count_sw / mean_count_lw
    else:
        mean_ratio = 0

    summary = {
        'key_date': date_index[-1].strftime("%Y-%m-%d"),
        'count': round(counts.sum(), 2),
        'count in trend window': round(counts_sw.sum(), 2),
        'mean count': float("%.2f" % round(mean_count_lw, 2)),
        'mean count in trend window': float("%.2f" % round(mean_count_sw, 2)),
        'mean ratio': round(mean_ratio, 2),
        # increase in % over the last 7 days using the smooth counts
        'trendinglast7days': calc_percent_change(counts, -1, -7),
        # increase in % over the last 30 days using the smooth counts
        'trendinglast30days': calc_percent_change(counts, -1, -30),
    }
    return summary


def update_timeline(data_df, date_index, smooth_factor):
    """
    on demand - get a list of complaint counts per day for this query
                get a list of complaints counts averaged over the last n days, n=smooth_factor
    create  a new data frame with Date as index and record count as the only column
    :return:
    :return:
    :param  select_expr:
    :param  data_df:
    :param  date_index: all the days in the long window in date format as an index
    :return: tuple: a list of complaint counts per day for this select expression, smoothed complaint_counts
    """
    if not data_df.empty:
        # re-index to have a consistent start and end date for all timelines
        df_timeline = df_timeline.reindex(index=date_index, fill_value=0)
        # rename to get the counts columns
        df_timeline.rename(columns={'ID': 'actualcount'}, inplace=True)

        df_timeline['smoothcount'] = np.round(df_timeline['actualcount'].rolling(window=smooth_factor).mean(), 2)
        df_timeline.fillna(0, inplace=True)

        return list(df_timeline['actualcount']), list(df_timeline['smoothcount'])
    else:  # no values found, return zero values
        return [0 for _ in date_index], [0 for _ in date_index]


def on_demand_timeline(select_expr, data_df, date_index, smooth_factor):
    """
    on demand - get a list of complaint counts per day for this query
                get a list of complaints counts averaged over the last n days, n=smooth_factor
    create  a new data frame with Date as index and record count as the only column
    :return:
    :return:
    :param  select_expr:
    :param  data_df:
    :param  date_index: all the days in the long window in date format as an index
    :param smooth_factor
    :return: tuple: a list of complaint counts per day for this select expression, smoothed complaint_counts
    """

    # filter data set on expression
    expr = ' & '.join([str(k) + ' == "' + str(v) + '"' for k, v in select_expr.items()])
    df_query = data_df.query(expr)

    if not df_query.empty:  # df_query = those rows in df-data that were selected by the query
        # groupby date and count per day
        cols = [k for k in select_expr.keys()]
        values = tuple(select_expr.values())
        df_timeline = df_query.set_index(cols).sort_index()
        df_timeline = df_timeline.loc[values, ['date', 'ID']].groupby('date').count()

        # re-index to have a consistent start and end date for all timelines
        df_timeline = df_timeline.reindex(index=date_index, fill_value=0)
        # rename to get the counts columns
        df_timeline.rename(columns={'ID': 'actualcount'}, inplace=True)

        df_timeline['smoothcount'] = np.round(df_timeline['actualcount'].rolling(window=smooth_factor).mean(), 2)
        df_timeline.fillna(0, inplace=True)

        return list(df_timeline['actualcount']), list(df_timeline['smoothcount'])
    else:  # no values found, return zero values
        return [0 for _ in date_index], [0 for _ in date_index]


def od_datecounts(time_period, counts_per_day, smth_cnts_per_day, date_index):
    """
    put counts per day in json format for the UI
    example {"complaintscount": 3, "date": "05/01/2017", "smoothcount": 3.0}
    :param time_period: ie last n days of this time period
    :param counts_per_day: raw counts
    :param smth_cnts_per_day:
    :param date_index: date for each count
    :return: list of counts : {"complaintscount": 3, "date": "05/01/2017", "smoothcount": 3.0}
    """
    df_timeline = pd.DataFrame({'complaintscount': counts_per_day, 'smoothcount': smth_cnts_per_day}, index=date_index)
    df_timeline.reset_index(inplace=True)
    df_timeline['date'] = df_timeline['date'].map(lambda x: x.strftime('%m/%d/%Y'))
    datecounts = df_timeline.iloc[-time_period:].to_dict(orient='records')
    return list(datecounts)


def analyze_trend_od(counts, date_index, trend_win_length, model_, mk_conf, poisson_threshold):
    """
    on demand analyze trend
    looks at a long window of counts and see if a trend is occurring in the trend window
    or last 7 days

    :param counts:      a list of raw counts over the long window
    :param date_index:
    :param trend_win_length: ie short window of 7 days
    :param model_:
    :param mk_conf mann-kendall confidence threshold
    :param poisson_threshold:
    :return: a dictionary of summary stats per trend
    """
    mk_dict = {'increasing': 1, 'decreasing': 2, 'no trend': 3}

    ts = np.asarray(counts[0])

    count_in_trend_window = ts[-trend_win_length:].sum()
    mean_count = ts.mean()
    mean_count_in_trend_window = ts[-trend_win_length:].mean()
    mean_ratio = 0
    if mean_count:
        mean_ratio = mean_count_in_trend_window / mean_count  # ratio of mean in trend window to mean of whole period

    # poisson model is used to identify exceptional high or low counts
    spd = 86400  # number of seconds in a day
    pm_values = [[datetime.strftime(d, '%Y-%m-%d'), s, c]
                 for (d, s, c) in zip(date_index, [spd for _ in range(len(ts))], counts[0])]

    pm_scores = td_analysis.analyze(pm_values, model_)[-trend_win_length:]
    poisson_scores_max = max([i[2] for i in pm_scores])
    poisson_above_thresh_count_inc = sum([1 for i in pm_scores if (i[2] > poisson_threshold) & (i[1] > mean_count)])
    poisson_above_thresh_count_dec = sum([1 for i in pm_scores if (i[2] > poisson_threshold) & (i[1] < mean_count)])

    moving_average = counts[1][-trend_win_length:]
    mann_kendall_score = td_models.mk_test(moving_average, mk_conf)

    # calculate the monotonicity score depending on trend type: 'decreasing':2, 'increasing':1, 'no trend':3
    monotone_score = 0
    if mann_kendall_score[0] == 'increasing':
        monotone_score = sum([x < y for x, y in zip(ts[-trend_win_length:], ts[-(trend_win_length - 1):])]) \
                         / (trend_win_length - 1)
    if mann_kendall_score[0] == 'decreasing':
        monotone_score = sum([x > y for x, y in zip(ts[-trend_win_length:], ts[-(trend_win_length - 1):])]) \
                         / (trend_win_length - 1)

    # output results to a dictionary
    summary = {
        'count': round(ts.sum(), 2),
        'count in trend window': round(count_in_trend_window, 2),
        'mean count': float("%.2f" % round(mean_count, 2)),
        'mean count in trend window': float("%.2f" % round(mean_count_in_trend_window, 2)),
        'mean ratio': round(mean_ratio, 2),
        # increase in % over the last 7 days using the smooth counts
        'trendinglast7days': calc_percent_change(counts[1], -1, -7),
        # increase in % over the last 30 days using the smooth counts
        'trendinglast30days': calc_percent_change(counts[1], -1, -30),
        'trendtypeid': mk_dict[mann_kendall_score[0]],  # 'decreasing':2, 'increasing':1, 'no trend':3
        'mk trend': mann_kendall_score[1],
        'mk score p-value': round(mann_kendall_score[2], 3),
        'poisson_scores_max': poisson_scores_max,
        'poisson_above_thresh_count_inc': poisson_above_thresh_count_inc,
        'poisson_above_thresh_count_dec': poisson_above_thresh_count_dec,
        'trend_detected': False,  # set default value
        'riskscore': 0.0,  # set default value
        'monotonicity': monotone_score
    }
    return summary


def create_complaint_count(row):
    """
    BATCH create UI complaint count field
    :param row:
    :return:
    """
    summary = row['summary']
    datecounts = row['datecounts']
    strt_date = row["trendstartdate"]

    row["complaintcount"] = {"dates": datecounts,
                             "7dayavg": summary['mean count in trend window'],
                             "30dayavg": summary['mean count'],
                             'trendinglast7days': summary['trendinglast7days'],
                             'trendinglast30days': summary['trendinglast30days'],
                             "earlytrend": strt_date
                             }
    return row


def create_complaint_count_od(row):
    """
    ON-DEMAND: create UI complaint count field
    :param row:
    :return:
    """
    summary = row['summary']
    datecounts = row['datecounts']
    strt_date = row["trendstart"]

    row["complaintcount"] = {"trendanalysis": datecounts,
                             "7dayavg": summary['mean count in trend window'],
                             "30dayavg": summary['mean count'],
                             'trendinglast7days': summary['trendinglast7days'],
                             'trendinglast30days': summary['trendinglast30days'],
                             "earlytrend": strt_date
                             }
    return row


def map_selections(selection_params):
    """
    turns selection parameters into a list of dicts
        {'THEME': 'Loan servicing, payments, escrow account', 'GEO': 'GA', 'PRODUCT': 'Mortgage'}
    is converted into
        [{'THEME': 'Loan servicing, payments, escrow account'}, {'GEO': 'GA'}, {'PRODUCT': 'Mortgage'}]

    also updates the selection criteria for age, and replaces it with a min and max age
        {'CUSTOMER_AGE': '20-29'} => {'MIN_AGE': 20}, {'MAX_AGE': 29}
    :param selection_params:
    :return:
    """
    select_params = selection_params.copy()

    if 'CUSTOMER_AGE' in select_params.keys():
        age_from, age_to = select_params['CUSTOMER_AGE'].split('-')
        select_params['MIN_AGE'] = int(age_from)
        select_params['MAX_AGE'] = int(age_to)
        del select_params['CUSTOMER_AGE']

    return [{k: v} for k, v in select_params.items() if v != 'NaN']


def map_selections_on_demand(selection_params):
    """
    turns selection parameters into a list of dicts
        {'THEME': 'Loan servicing, payments, escrow account', 'GEO': 'GA', 'PRODUCT': 'Mortgage'}
    is converted into
        [{'THEME': 'Loan servicing, payments, escrow account'}, {'GEO': 'GA'}, {'PRODUCT': 'Mortgage'}]

    also updates the selection criteria for age, and replaces it with a min and max age
        {'CUSTOMER_AGE': '20-29'} => {'MIN_AGE': 20}, {'MAX_AGE': 29}
    :param selection_params:
    :return:
    """
    params_formatted = []
    select_params = selection_params.copy()
    for k, v in select_params.items():
        if k == 'CUSTOMER_AGE':

            age_ranges = [age_boundary(age_range) for age_range in select_params['CUSTOMER_AGE']]
            age_ranges = [tuple(age_range.split('-')) for age_range in age_ranges]
            age_ranges = sorted(age_ranges, key=lambda x: x[0])
            params_formatted.append({'MIN_AGE': int(age_ranges[0][0])})
            params_formatted.append({'MAX_AGE': int(age_ranges[-1][1])})
        # elif 'DATE' in k:
        #     params_formatted.append({k: v})
        elif v != 'NaN':
            params_formatted.append({k: ', '.join(v)})

    return params_formatted


def get_age(selection_params):
    """
    return formatted age range:     {'CUSTOMER_AGE': '20-30', '30-40'} => "20-40"
    :param selection_params:
    :return:
    """
    ages = ""
    for k, v in selection_params.items():
        if k == 'CUSTOMER_AGE':
            age_ranges = [age_boundary(age_range) for age_range in selection_params['CUSTOMER_AGE']]
            age_ranges = [tuple(age_range.split('-')) for age_range in age_ranges]

            age_ranges = sorted(age_ranges, key=lambda x: x[0])
            ages = '{}-{}'.format(age_ranges[0][0], age_ranges[-1][1])
    return ages


def format_dates(date_str):
    """
    convert various different date formats into a datetime field
    :param date_str:
    :return:
    """
    fmts = ('%m/%d/%Y', '%Y-%m-%d', '%d/%m/%Y', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S.%f')
    date_fmt = ''
    for fmt in fmts:
        try:
            date_fmt = datetime.strptime(date_str, fmt)
            break
        except ValueError:
            pass

    return date_fmt


def calc_riskscore(summary_dict, risk_max, count_max, risk_wts, grad_max):
    """ calculates the risk score for each detected trend
    score = (normalised_count ** count_weight)
            * (normalised_gradient ** grad_weight)
            * (monotone_score ** monotone_weight)

    :param summary_dict: dictionary of summary statistics
    :param risk_max: maximum possible risk score eg 85%
    :param count_max: max count expected in a trend
    :param risk_wts: list of risk weights for count, gradient, and monotonicity respectively
    :param grad_max: max gradient expected in a trend
    :return: a risk score %
    """

    log_max_count = math.log10(count_max)
    log_max_grad = math.log10(grad_max)
    # count is thresholded at the max level and then normalised to give a value 0 - 1
    if summary_dict['count_in_trend'] > 0:
        count_norm = min(math.log10(summary_dict['count_in_trend']), log_max_count) / log_max_count
    else:
        count_norm = 0
    # gradient is thresholded at the max level and then and then normalised to give a value 0 - 1
    grad_norm = min(math.log10(abs(summary_dict['gradient'])), log_max_grad) / log_max_grad
    # monotonicity is already normalised to give a value 0 - 1
    monotone_norm = summary_dict['monotonicity']

    score = count_norm ** risk_wts[0] * grad_norm ** risk_wts[1] * monotone_norm ** risk_wts[2]

    # threshold the value between 0 and 100 and scale by max risk (85%)
    return float("%.2f" % (risk_max * max(min(score, 100.0), 0.0) / 100.0))


def is_true(cmplnt):
    if isinstance(cmplnt, bool):
        return cmplnt
    elif isinstance(cmplnt, str):
        return 'TRUE' in cmplnt or 'True' in cmplnt
    else:
        return cmplnt == 1


def check_num(zc):
    """create a State column by looking up the zipcode, this checks if the zip code is a valid string
    :param zc: zip code or blank or string
    :return: zc or 0 if x is not numeric
    """
    try:
        ret_code = int(zc)
    except ValueError:
        ret_code = 0
    return ret_code


def post_request(trends_, url, vfy):
    """
    send the found trends to a rest api
    :param trends_:
    :param trends_: json structure of trends
    :param url:
    :param vfy:
    :return: the request response
    """
    params = dict()  # dict for the params
    params['complainttrends'] = trends_  # param1 = detected trends send as a list
    return requests.post(url, json=params, verify=vfy)


def merge_tuples(xs):
    """
    given a list of age ranges, this will merge the ranges if consecutive
    example:
    age_ranges in = [ '40-50', '30-40']
    age_range out = [(30,50)]
    :param xs: list of age range strings in format 'from-to'
    :return: tuple(s) of merged age ranges
    """
    it = iter(xs)
    ret = next(it, (None, None))
    for x in it:
        if ret[1] is None or x[0] != ret[1]:
            yield ret
            ret = x
        else:
            ret = ret[0], x[1]
    if ret[0] is not None:
        yield ret


def get_risk(themes, riskmap):
    """
    return the mapping from theme-> risk eg Delay = High
    :param themes:
    :param riskmap
    :return: risk rating for this theme: HIGH, MEDIUM or LOW
    """
    themes = [t.strip() for t in themes.split(',')]   # split up themes from a string

    risks = [riskmap.get(t, "") for t in themes]           # get risk for each

    if 'HIGH' in risks:                               # only return the highest value
        return 'HIGH'
    elif 'MEDIUM' in risks:
        return 'MEDIUM'
    elif 'LOW' in risks:
        return ['LOW']
    else:
        return 'N/A'


def early_trend(trendstartdate, counts, dt_index):
    """
    find the smooth count on the date the trending started
    :param trendstartdate: date the trending started
    :param counts: list of counts per day
    :param dt_index: datetimeindex for the timeseries
    :return: the count on the trend start date
    """
    count_on_ts = 0
    if trendstartdate:
        if datetime.strptime(trendstartdate, '%Y-%m-%d') in dt_index:
            count_on_ts = counts[dt_index.get_loc(trendstartdate)]
    return {'complaintscount': count_on_ts, 'date': trendstartdate}


def age_boundary(age_range):
    """
    turn age boundary clause eg <30 into a range 0-30
    :param age_range:
    :return:
    """
    if '<' in age_range:
        age_range = '0-' + age_range.split('<')[1]
    if '>' in age_range:
        age_range =  age_range.split('>')[1] + '-100'
    return age_range


def get_where_clause(sqlquery, start_date, end_date):
    """
    create an sql where clause for  select statement
    1) any select fields
    2) the age ranges
    3) start and end date
    :param sqlquery:
    :param start_date:
    :param end_date:
    :return:
    """
    where_s = []
    query_fields_list = [str(k) for k in sqlquery.keys()]

    for field, values in sqlquery.items():
        if 'AGE' not in field and 'DATE' not in field:
            if len(values) == 1:
                where_s.append(str(field) + " = '" + str(values[0]) + "'")
            else:
                where_s.append(str(field) + " in (" + ' ,'.join(["'" + str(v) + "'" for v in values]) + ')')

    # create a range select for age
    if 'CUSTOMER_AGE' in query_fields_list:
        age_list = sqlquery.get('CUSTOMER_AGE')
        # change <30 to 0-30, and >60 to 60-100
        age_list = [age_boundary(age) for age in age_list]

        if len(age_list) == 1:
            from_age, to_age = tuple(age_list[0].split('-'))
            # update ages
            age_s = 'CUSTOMER_AGE BETWEEN {} and {}'.format(from_age, to_age)

        else:  # multiple age criteria
            age_ranges = [tuple(age_range.split('-')) for age_range in age_list]
            age_ranges = sorted(age_ranges, key=lambda x: x[0])
            age_ranges_merged = list(merge_tuples(age_ranges))
            if len(age_ranges_merged) == 1:
                age_s = 'CUSTOMER_AGE BETWEEN {} and {}'.format(*age_ranges_merged[0])
            else:
                age_s = ' or '.join(['CUSTOMER_AGE BETWEEN {} and {}'.format(*t) for t in age_ranges_merged])
                age_s = '( ' + age_s + ')'

        # append results to "where" string
        where_s.append(age_s)

    # add start and end date to query
    sd_s = datetime.strptime(start_date, '%Y-%m-%d')
    ed_s = datetime.strptime(end_date, '%Y-%m-%d')
    sd = datetime.strftime(sd_s, '%m/%d/%Y')
    ed = datetime.strftime(ed_s, '%m/%d/%Y')

    date_s = "DATE(COMPLAINT_DATE) between DATE('{}') and DATE('{}')".format(sd, ed)
    where_s.append(date_s)

    return ' and '.join(where_s)


def get_source(od_raw_counts):
    """
    get source counts function placeholder
    :param od_raw_counts:
    :return: list of dicts  [{"count": "246", "name": "Email"}]
    """
    source = {}
    source['name'] = "Email"
    source['count'] = str(sum(od_raw_counts))  # has to be a string, cant be an integer
    return [source]  # return as a list


def get_config(sifs_config_file, config_file, sslServerCert, logging):
    """
    get all configuration required by the trends detection algorithm
    :param sifs_config_file:
    :param config_file:
    :param sslServerCert:
    :param logging:
    :return:
    """

    if not os.path.isfile(sifs_config_file):
        txt = 'Unable to find global SIFS config file: ' + sifs_config_file + '. Update file path or use parameter' \
                                                                              '-t to read local file ../config/sifs.spark.properties.  Ending Script.'
        logging.error(txt)
        raise ValueError(txt)

    lines = open(sifs_config_file).readlines()
    lines = [l for l in lines if l[0] != '#']

    # read db2 integration params
    db2_params = {'db2user': '', 'db2password': '', 'db2jdbcurl': '', 'db2TrustStore': ''}
    for p in db2_params.keys():
        param = [l for l in lines if str(p) in l]
        if param:
            p_name, p_val = param[0].split('=')[0], '='.join(param[0].split('=')[1:])
            db2_params[p] = p_val.replace('\n', '').strip()
            logging.info('Read parameter {} from config file: {}'.format(p, db2_params[p]))
        else:
            logging.error('Unable to find parameter {} in config file. Ending Script.'.format(p))
            quit()

    # change this to decode from base64
    if db2_params['db2password'].__contains__('{xor}'):
        db2_params['db2password'] = db2_params['db2password'][5:]
        db2_params['db2password'] = base64.b64decode(db2_params['db2password'])
        db2_params['db2password'] = bytearray(i ^ ord('_') for i in db2_params['db2password'])
        db2_params['db2password'] = db2_params['db2password'].decode('utf-8')
    else:
        logging.error("XOR-base64 Encoding not found in config")
        quit()
        
    # get db
    match = re.search(r"(?<=currentSchema=)\w*", db2_params['db2jdbcurl'])
    if match:
        db2_params['database'] = match.group()
    # get ssl flag
    match = re.search(r"(?<=sslConnection=)\w*", db2_params['db2jdbcurl'])
    if match:
        db2_params['sslConnection'] = match.group()

    # get host ip and port from db2 url
    if '//' in db2_params['db2jdbcurl']:
        host_port = db2_params['db2jdbcurl'].split("//")[-1].split("/")[0].split(":")
        if len(host_port) == 2:
            db2_params['host'] = host_port[0]
            logging.info('Read db2 host ip address from config file: {}'.format(db2_params['host']))
            db2_params['port'] = host_port[1]
            logging.info('Read db2 port from config file: {}'.format(db2_params['port']))
        else:
            txt = 'Unable to parse db2 url: {}. Ending Script'.format(db2_params['db2jdbcurl'])
            logging.error(txt)
            raise ValueError(txt)
    else:
        txt = 'Unable to parse db2 url: {}. Ending Script'.format(db2_params['db2jdbcurl'])
        logging.error(txt)
        raise ValueError(txt)

    # temp fix store public certificate in REST config file, needs a better home
    db2_params['sslServerCert'] = sslServerCert  # this will get updated, expect it to end up in db2params
    if not os.path.isfile(db2_params['sslServerCert']):
        txt = 'Unable to find db2 SSL Certificate: ' + db2_params['sslServerCert'] + '. Ending Script.'
        logging.error(txt)
        raise ValueError(txt)

    # concat params into a db2 connection string for a secure (ssl) server
    db2_params['conn_str'] = "DATABASE=%s;HOSTNAME=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s;" \
                             "SECURITY=SSL;SSLServerCertificate=%s;" % \
                             (db2_params['database'],
                             db2_params['host'],
                             db2_params['port'],
                             db2_params['db2user'],
                             db2_params['db2password'],
                             db2_params['sslServerCert'])

    config = configparser.ConfigParser()
    if config_file is None:
        logging.error('No trends config file parameter given. Ending Script.')
        raise ValueError('No trends config file parameter given. Ending Script.')
    else:
        # read config file
        cfg = config.read(config_file)
        if not len(cfg):
            txt = 'Unable to read config file or blank config file given. Ending Script.' + config_file
            logging.error(txt)
            raise ValueError(txt)
        else:
            logging.info('Config files read completed')

    # establish connection to db2
    logging.info('connecting to host {} port {}'.format(db2_params['host'], db2_params['port']))
    try:
        ibm_db_conn = ibm_db.pconnect(db2_params['conn_str'], '', '')
        connection = ibm_db_dbi.Connection(ibm_db_conn)
        connection.set_current_schema(db2_params['database'])
        logging.info('connected to db2')
    except:
        txt = 'Unable to connect to db2'
        logging.error(txt)
        raise ValueError(txt)

    # add theme_risk_map
    qry = "select * from SIFS.THEME_RISK_MAP"
    try:
        df_risk = pd.read_sql(qry, connection, index_col='THEME')
    except:
        txt = "Unable to read SIFS.THEME_RISK_MAP, error running query: '{}'".format(qry)
        logging.error(txt)
        raise ValueError(txt)
    logging.info('finished reading the riskmap table')

    riskmap = df_risk.to_dict()['RISK_LEVEL']  # turn into a dict
    config.add_section('riskmap')                          # add it to the conf object
    config.set('riskmap', 'riskmap', json.dumps(riskmap))  # store as a string for now

    return db2_params, config


def get_trends(query, conn, config, logging):
    """
    main on-demand trend function
    :param query:
    :param conn:
    :param config:
    :param logging:
    :return:
    """

    # list of fields for the final json structure
    keep_cols = ['description', 'topgeography', 'product', 'theme', 'trendstart', 'trendanalysis',
                 'risklevel', 'totalcomplaints', 'agegroup', 'earlytrend', 'sources']

    # read trend detection model parameters
    table = config.get("db2 integration", "table")  # table = SIFS.COMPLAINT_FEATURES
    # trend detection params
    mk_confidence = config.getfloat("mann kendall", "mk_confidence")
    min_records = config.getint("trend candidates", "min_records")
    min_per_day = config.getint("trend candidates", "min_per_day")
    pct_change_threshold = config.getfloat("trend candidates", "pct_change_threshold")
    MIN_MONOTONE = config.getfloat("trend candidates", "min_monotone")
    fields_list = config.get("trend candidates", "fields_list")
    fields_list = fields_list.replace('-', '_').split(',')
    products = config.get("trend candidates", "products")
    model_config = dict(config.items('poisson model'))
    model_name = config.get('poisson model', "model_name")
    MAX_RISK = config.getfloat("risk score", "max_risk")  # 85% # 503 max risk score allowable is 85%
    RISK_WEIGHTS = json.loads(config.get("risk score", "risk_weights"))  # rel. factor weights for scoring trends
    MAX_COUNT = config.getfloat("risk score", "max_count")
    MAX_GRAD = config.getfloat("risk score", "max_grad")
    MIN_RISK_FILTER = config.getfloat("risk score", "min_risk_filter")

    # on-demand params
    short_window = config.getint("on-demand", "short_window")
    long_window = config.getint("on-demand", "long_window")
    # number of days for averaging the raw complaint count. e.g. use 7 day average to remove weekly seasonality
    smoothfactor = config.getint("on-demand", "smoothfactor")
    # trend on days less than this factor apart will be joined up to find the earliest date the trend started
    DAYS_BETWEEN_TRENDS = config.getint("on-demand", "DAYS_BETWEEN_TRENDS")
    # number of time periods to calculate the on-demand timelines
    time_periods = json.loads(config.get("on-demand", "time_periods"))  # e.g. [365, 90, 30, 7]

    # check for filter on products
    if products == 'all':
        products = []
    else:
        products = products.split(",")

    # check validity of risk weights
    if len(RISK_WEIGHTS) != 3:
        # warning default equal weighting used for all weights
        txt = 'Default values are used to give equal weighting to all factors in the risk score calculation.'
        logging.info(txt)
        RISK_WEIGHTS = [1 / 3 for _ in range(3)]
    # normalise weights if required
    if sum(RISK_WEIGHTS) != 1:
        RISK_WEIGHTS = np.asarray(RISK_WEIGHTS) / sum(RISK_WEIGHTS)

    # load the theme-> risk level mapping as a dictionary
    riskmap = json.loads(config.get('riskmap', 'riskmap'))

    # initialise a poisson trend detection model
    model = getattr(td_models, model_name)(config=model_config)

    # dynamically build up the query
    logging.info('evaluate UI query: {}'.format(json.dumps(query)))

    # check for REF fields:
    #  its possible that GEO is a being queried but geo data is actually stored in the GEO REF column
    sql_query = query.copy()  # sql_query is used to build the query, query is used for outputting texts etc
    sql_query.update({'IS_COMPLAINT': ['1']})  # we only want complaints, so add this to where clause in sql query
    ref_fields = ['PRODUCT', 'SUB_PRODUCT', 'GEO']
    replace_fields = [rf for rf in ref_fields if (rf in query) and (rf + '_REF' in fields_list)]
    for rf in replace_fields:
        sql_query[rf + '_REF'] = query[rf]
        del sql_query[rf]

    if replace_fields:
        logging.info('query updated to use _REF fields: {}'.format(json.dumps(sql_query)))

    # get start date and end date
    end_date = query.get('ENDDATE', None)
    if end_date:
        if valid_date(end_date):
            logging.info('end date parameter: {}'.format(end_date))
            query['ENDDATE'] = [end_date]
        else:
            txt = 'Invalid end date parameter: {}. Ending Script.'.format(end_date)
            logging.error(txt)
            raise ValueError(txt)
    else:
        logging.info('end date: Parameter "ENDDATE" not in query, using todays date.')
        end_date = datetime.now().strftime("%Y-%m-%d")
    dt_end_date = datetime.strptime(end_date, '%Y-%m-%d')

    start_date = query.get('STARTDATE', None)
    if start_date:
        if valid_date(start_date):
            logging.info('start date parameter: {}'.format(start_date))
            query['STARTDATE'] = [start_date]
        else:
            txt = 'Invalid start date parameter: {}. Ending Script.'.format(start_date)
            logging.error(txt)
            raise ValueError(txt)
    else:
        logging.info('start date: start today, go back {} days + use prior {} days to get context'
                     .format(max(time_periods) - 1, long_window))
        # the start date is the greatest number of days in time periods + the number of days in the long window
        start_date = datetime.strftime(dt_end_date - timedelta(days=long_window + max(time_periods) - 1), '%Y-%m-%d')

    logging.info('on-demand date range: start date={}'.format(start_date))
    logging.info('on-demand date range: end   date={}'.format(end_date))

    # dynamically build up the query
    # update after changing db2 column to timestamp: select CHAR(DATE(COMPLAINT_DATE), USA)

    select_fields = ["CHAR(DATE(COMPLAINT_DATE), USA) as date",
                     'COUNT(ID) as count']  # need these to build a timeline
    select_fields_s = ' ,'.join(select_fields)
    where_s = get_where_clause(sql_query, start_date, end_date)
    groupby_s = 'CHAR(DATE(COMPLAINT_DATE), USA)'

    qry = 'select {} from {} where {} group by {}'.format(select_fields_s, table, where_s, groupby_s)
    logging.info('created main db2 query: {}'.format(json.dumps(qry)))

    # run the query
    logging.info('started reading the data')
    try:
        df_data = pd.read_sql(qry, conn, index_col='DATE')
    except:
        txt = "error running query: '{}'".format(qry)
        logging.error(txt)
        raise ValueError(txt)
    logging.info('finished reading the data')

    # 2nd query to derive source of complaints for UI
    # example {"sources": [{"count": 68, "name": "Email"}]
    # #1702 also theme
    select_fields = ['distinct(SOURCE) as name', 'count(Source) as count, theme']
    select_fields_s = ', '.join(select_fields)
    qry = 'select {} from {} where {} group by SOURCE, THEME'.format(select_fields_s, table, where_s)
    logging.info('created source db2 query: {}'.format(json.dumps(qry)))
    try:
        df_source = pd.read_sql(qry, conn)
    except:
        txt = "error running sources query: '{}'".format(qry)
        logging.error(txt)
        raise ValueError(txt)
    logging.info('finished reading the complaint sources')

    # extract the themes and then drop these from the df
    themes = df_source['THEME'].values
    themes = [t for t in themes if t in riskmap.keys()]
    df_source.drop('THEME', inplace=True, axis=1)
    df_source = df_source.groupby("NAME").sum()
    df_source.reset_index(inplace=True, drop=False)
    df_source.rename(columns={'NAME': 'name', 'COUNT': 'count'}, inplace=True)
    sources = df_source.to_dict(orient='rows')  # convert to list of k:v pairs

    # df_data has string index with US date format, change this to datetime index in YYYY-mm-dd
    df_data.index = [datetime.strptime(d, '%m/%d/%Y') for d in df_data.index.values]
    od_index = pd.date_range(start=start_date, end=end_date, freq='1D', name='date')
    df_data = df_data.reindex(index=od_index, fill_value=0)
    df_data.rename(columns={'COUNT': 'actualcount'}, inplace=True)
    df_data['smoothcount'] = np.round(exp_smooth(df_data['actualcount'].values), 2)
    df_data.fillna(0, inplace=True)
    od_raw_counts, od_smooth_counts = list(df_data['actualcount']), list(df_data['smoothcount'])

    # set defaults and initialisations
    summary = None
    trends = []
    trendstartdate = ""

    # long window ( eg 30 days) is used to create an average baseline for comparison
    # check that this amount is not longer than the total timespan, if so run on the last day only
    if long_window >= len(od_index):
        lw_temp = long_window
        long_window = len(od_index) - 2
        print('resetting long window from {} to {} days'.format(lw_temp, len(od_index) - 2))
        logging.info('resetting long window from {} to {} days'.format(lw_temp, len(od_index) - 2))

    for day in od_index[long_window:]:  # run a sliding window of length <<long_window>> over the whole period
        # index of "to" date, add a +1 because want results to be inclusive of this end date
        idx_to = list(od_index).index(day) + 1
        # index of "from" date
        idx_from = idx_to - long_window
        # return one row per trend discovered
        trend = od_get_trends(od_raw_counts[idx_from: idx_to], od_smooth_counts[idx_from: idx_to],
                              od_index[idx_from: idx_to], short_window, long_window, MIN_MONOTONE, model,
                              mk_confidence, float(model_config['poisson_threshold']), pct_change_threshold,
                              MAX_RISK, MAX_COUNT, RISK_WEIGHTS, MAX_GRAD, min_records, min_per_day)
        if trend:
            trends.append(trend)

        # derive a minimal set of timeline statistics
        # if no trends have been found, still show trending 30 days etc so user can see "no trend" is correct analysis
        if day == od_index[-1]:
            summary_min = get_min_stats(od_raw_counts[idx_from: idx_to], od_index[idx_from: idx_to], short_window,
                                        long_window)

    # trend details: use the trend summary from the end date
    logging.info('Found {} trending days for this query'.format(len(trends)))

    for trend in sorted(trends, key=lambda x: x['key_date'], reverse=True):
        if trend['key_date'] == end_date:             # was this timeseries trending at the end
            logging.info('Query was trending on {}. The trend started on {}'.format(end_date, trend['trendstartdate']))
            summary = trend                           # if so set this as a trend
            trendstartdate = trend['trendstartdate']
            key_date = end_date

        elif trendstartdate:  # if trend startdate already set-> see if it started before the last day
            date_diff = datetime.strptime(key_date, '%Y-%m-%d') - datetime.strptime(trend['key_date'], '%Y-%m-%d')
            if date_diff.days <= DAYS_BETWEEN_TRENDS:
                logging.info('Found earlier trend within {} days. New trend start date is {}.'.format
                             (DAYS_BETWEEN_TRENDS, trend['trendstartdate']))
                trendstartdate = trend['trendstartdate']  # overwrite with earlier trend start date
                key_date = trend['key_date']

    # if no trend has been found then give a minimal set of timeline stats
    if not summary:
        summary = summary_min

    description = "&".join([k.replace('&', '_') + ':' + ','.join([str(v).replace('&', '_')
                                                                  for v in vs]) for k, vs in query.items()])

    # create dataframe with common values
    df_trends_od = pd.DataFrame({'selection_criteria': [query for _ in time_periods],
                                 'timeperiod': sorted(time_periods, reverse=True),
                                 'description': [description for _ in time_periods],
                                 'product': [', '.join(query.get('PRODUCT', ['all'])) for _ in time_periods],
                                 'trendattributes': [map_selections_on_demand(query) for _ in time_periods],
                                 'topgeography': [','.join(query.get('GEO', ['all'])) for _ in time_periods],
                                 'theme': [', '.join(themes) for _ in time_periods],  # 1702 update : themes from query
                                 'summary': [summary for _ in time_periods],
                                 'trendstart': [trendstartdate for _ in time_periods]})

    df_trends_od['datecounts'] = df_trends_od['timeperiod'].apply(od_datecounts, args=(od_raw_counts,
                                                                                       od_smooth_counts,
                                                                                       od_index))

    df_trends_od['date_range'] = df_trends_od['timeperiod'].apply(lambda x: datetime.strftime(dt_end_date -
                                                                                              timedelta(days=x),
                                                                                              '%Y-%m-%d') + ':' +
                                                                                               end_date)

    df_trends_od["score"] = df_trends_od['summary'].map(lambda x: x.get('riskscore', 0.0))
    df_trends_od = df_trends_od.apply(create_complaint_count_od, axis=1)
    df_trends_od["trendanalysis"] = df_trends_od['complaintcount'].map(lambda x: x.get('trendanalysis'))
    # new fields for #
    df_trends_od['risklevel'] = df_trends_od['theme'].apply(get_risk, args=(riskmap,))  # risk level for this trend
    df_trends_od['totalcomplaints'] = sum(od_raw_counts)                                # total count of complaints
    df_trends_od['agegroup'] = [get_age(query) for _ in time_periods]
    df_trends_od['earlytrend'] = df_trends_od['trendstart'].apply(early_trend, args=(od_smooth_counts, od_index))
    df_trends_od['sources'] = [sources for _ in time_periods]

    #  append results to output structure
    od_trends = df_trends_od[keep_cols].to_dict(orient='records')

    if len(od_trends) == 1:
        return od_trends[0]  # return either a dict
    else:
        return od_trends     # or list of dicts
